import logging
import re
from copy import deepcopy
from datetime import timedelta
from typing import Any, Dict, List, Optional, Union

from pydantic import (
    Field,
    PositiveInt,
    PrivateAttr,
    ValidationInfo,
    field_validator,
    model_validator,
)

from datahub.configuration.common import AllowDenyPattern, ConfigModel, HiddenFromDocs
from datahub.configuration.env_vars import get_bigquery_schema_parallelism
from datahub.configuration.source_common import (
    EnvConfigMixin,
    LowerCaseDatasetUrnConfigMixin,
    PlatformInstanceConfigMixin,
)
from datahub.configuration.validate_field_removal import pydantic_removed_field
from datahub.ingestion.glossary.classification_mixin import (
    ClassificationSourceConfigMixin,
)
from datahub.ingestion.source.bigquery_v2.bigquery_connection import (
    BigQueryConnectionConfig,
)
from datahub.ingestion.source.data_lake_common.path_spec import PathSpec
from datahub.ingestion.source.sql.sql_config import SQLCommonConfig, SQLFilterConfig
from datahub.ingestion.source.state.stateful_ingestion_base import (
    StatefulLineageConfigMixin,
    StatefulProfilingConfigMixin,
    StatefulTimeWindowConfigMixin,
    StatefulUsageConfigMixin,
)
from datahub.ingestion.source.usage.usage_common import BaseUsageConfig

logger = logging.getLogger(__name__)

DEFAULT_BQ_SCHEMA_PARALLELISM = get_bigquery_schema_parallelism()

# Regexp for sharded tables.
# A sharded table is a table that has a suffix of the form _yyyymmdd or yyyymmdd, where yyyymmdd is a date.
# The regexp checks for valid dates in the suffix (e.g. 20200101, 20200229, 20201231) and if the date is not valid
# then it is not a sharded table.
_BIGQUERY_DEFAULT_SHARDED_TABLE_REGEX: str = (
    "((.+\\D)[_$]?)?(\\d\\d\\d\\d(?:0[1-9]|1[0-2])(?:0[1-9]|[12][0-9]|3[01]))$"
)


class BigQueryBaseConfig(ConfigModel):
    rate_limit: bool = Field(
        default=False, description="Should we rate limit requests made to API."
    )
    requests_per_min: int = Field(
        default=60,
        description="Used to control number of API calls made per min. Only used when `rate_limit` is set to `True`.",
    )

    temp_table_dataset_prefix: str = Field(
        default="_",
        description="If you are creating temp tables in a dataset with a particular prefix you can use this config to set the prefix for the dataset. This is to support workflows from before bigquery's introduction of temp tables. By default we use `_` because of datasets that begin with an underscore are hidden by default https://cloud.google.com/bigquery/docs/datasets#dataset-naming.",
    )

    sharded_table_pattern: str = Field(
        deprecated=True,
        default=_BIGQUERY_DEFAULT_SHARDED_TABLE_REGEX,
        description="The regex pattern to match sharded tables and group as one table. This is a very low level config parameter, only change if you know what you are doing, ",
    )

    @field_validator("sharded_table_pattern", mode="after")
    @classmethod
    def sharded_table_pattern_is_a_valid_regexp(cls, v: str) -> str:
        try:
            re.compile(v)
        except Exception as e:
            raise ValueError(
                "sharded_table_pattern configuration pattern is invalid."
            ) from e
        return v

    @model_validator(mode="before")
    @classmethod
    def project_id_backward_compatibility_configs_set(cls, values: Dict) -> Dict:
        # Create a copy to avoid modifying the input dictionary, preventing state contamination in tests
        values = deepcopy(values)
        project_id = values.pop("project_id", None)
        project_ids = values.get("project_ids")

        if not project_ids and project_id:
            values["project_ids"] = [project_id]
        elif project_ids and project_id:
            logging.warning(
                "Please use `project_ids` config. Config `project_id` will be ignored."
            )
        return values


class BigQueryUsageConfig(BaseUsageConfig):
    _query_log_delay_removed = pydantic_removed_field(
        "query_log_delay", month="April", year=2023
    )

    max_query_duration: timedelta = Field(
        default=timedelta(minutes=15),
        description="Correction to pad start_time and end_time with. For handling the case where the read happens "
        "within our time range but the query completion event is delayed and happens after the configured"
        " end time.",
    )

    apply_view_usage_to_tables: bool = Field(
        default=False,
        description="Whether to apply view's usage to its base tables. If set to False, uses sql parser and applies "
        "usage to views / tables mentioned in the query. If set to True, usage is applied to base tables "
        "only.",
    )


class GcsLineageProviderConfig(ConfigModel):
    """
    Any source that produces gcs lineage from/to Datasets should inherit this class.
    """

    path_specs: List[PathSpec] = Field(
        default=[],
        description="List of PathSpec. See below the details about PathSpec",
    )

    strip_urls: bool = Field(
        default=True,
        description="Strip filename from gcs url. It only applies if path_specs are not specified.",
    )

    ignore_non_path_spec_path: bool = Field(
        default=False,
        description="Ignore paths that are not match in path_specs. It only applies if path_specs are specified.",
    )


class GcsDatasetLineageProviderConfigBase(ConfigModel):
    """
    Any source that produces gcs lineage from/to Datasets should inherit this class.
    This is needeed to group all lineage related configs under `gcs_lineage_config` config property.
    """

    gcs_lineage_config: GcsLineageProviderConfig = Field(
        default=GcsLineageProviderConfig(),
        description="Common config for gcs lineage generation",
    )


class BigQueryFilterConfig(SQLFilterConfig):
    project_ids: List[str] = Field(
        default_factory=list,
        description=(
            "Ingests specified project_ids. Use this property if you want to specify what projects to ingest or "
            "don't want to give project resourcemanager.projects.list to your service account. "
            "Overrides `project_id_pattern`."
        ),
    )
    project_labels: List[str] = Field(
        default_factory=list,
        description=(
            "Ingests projects with the specified labels. Set value in the format of `key:value`. Use this property to "
            "define which projects to ingest based"
            "on project-level labels. If project_ids or project_id is set, this configuration has no effect. The "
            "ingestion process filters projects by label first, and then applies the project_id_pattern."
        ),
    )

    project_id_pattern: AllowDenyPattern = Field(
        default=AllowDenyPattern.allow_all(),
        description="Regex patterns for project_id to filter in ingestion.",
    )

    dataset_pattern: AllowDenyPattern = Field(
        default=AllowDenyPattern.allow_all(),
        description="Regex patterns for dataset to filter in ingestion. Specify regex to only match the schema name. "
        "e.g. to match all tables in schema analytics, use the regex 'analytics'",
    )

    match_fully_qualified_names: bool = Field(
        default=True,
        description="[deprecated] Whether `dataset_pattern` is matched against fully qualified dataset name "
        "`<project_id>.<dataset_name>`.",
    )

    table_snapshot_pattern: AllowDenyPattern = Field(
        default=AllowDenyPattern.allow_all(),
        description="Regex patterns for table snapshots to filter in ingestion. Specify regex to match the entire "
        "snapshot name in database.schema.snapshot format. e.g. to match all snapshots starting with "
        "customer in Customer database and public schema, use the regex 'Customer.public.customer.*'",
    )

    # NOTE: `schema_pattern` is added here only to hide it from docs.
    schema_pattern: HiddenFromDocs[AllowDenyPattern] = Field(
        default=AllowDenyPattern.allow_all(),
    )

    @model_validator(mode="after")
    def backward_compatibility_configs_set(self) -> Any:
        dataset_pattern = self.dataset_pattern
        schema_pattern = self.schema_pattern

        if (
            dataset_pattern == AllowDenyPattern.allow_all()
            and schema_pattern != AllowDenyPattern.allow_all()
        ):
            logging.warning(
                "dataset_pattern is not set but schema_pattern is set, using schema_pattern as dataset_pattern. "
                "schema_pattern will be deprecated, please use dataset_pattern instead."
            )
            self.dataset_pattern = schema_pattern
            dataset_pattern = schema_pattern
        elif (
            dataset_pattern != AllowDenyPattern.allow_all()
            and schema_pattern != AllowDenyPattern.allow_all()
        ):
            logging.warning(
                "schema_pattern will be ignored in favour of dataset_pattern. schema_pattern will be deprecated,"
                " please use dataset_pattern only."
            )

        match_fully_qualified_names = self.match_fully_qualified_names

        if (
            dataset_pattern is not None
            and dataset_pattern != AllowDenyPattern.allow_all()
            and match_fully_qualified_names is not None
            and not match_fully_qualified_names
        ):
            logger.warning(
                "Please update `dataset_pattern` to match against fully qualified schema name "
                "`<project_id>.<dataset_name>` and set config `match_fully_qualified_names : True`."
                "The config option `match_fully_qualified_names` is deprecated and will be "
                "removed in a future release."
            )
        elif match_fully_qualified_names and dataset_pattern is not None:
            adjusted = False
            for lst in [dataset_pattern.allow, dataset_pattern.deny]:
                for i, pattern in enumerate(lst):
                    if "." not in pattern:
                        if pattern.startswith("^"):
                            lst[i] = r"^.*\." + pattern[1:]
                        else:
                            lst[i] = r".*\." + pattern
                        adjusted = True
            if adjusted:
                logger.warning(
                    "`dataset_pattern` was adjusted to match against fully qualified schema names,"
                    " of the form `<project_id>.<dataset_name>`."
                )

        return self


class BigQueryIdentifierConfig(
    PlatformInstanceConfigMixin, EnvConfigMixin, LowerCaseDatasetUrnConfigMixin
):
    include_data_platform_instance: bool = Field(
        default=False,
        description="Whether to create a DataPlatformInstance aspect, equal to the BigQuery project id."
        " If enabled, will cause redundancy in the browse path for BigQuery entities in the UI,"
        " because the project id is represented as the top-level container.",
    )

    enable_legacy_sharded_table_support: bool = Field(
        default=True,
        description="Use the legacy sharded table urn suffix added.",
    )


class BigQueryV2Config(
    GcsDatasetLineageProviderConfigBase,
    BigQueryConnectionConfig,
    BigQueryBaseConfig,
    BigQueryFilterConfig,
    # BigQueryFilterConfig must come before (higher precedence) the SQLCommon config, so that the documentation overrides are applied.
    BigQueryIdentifierConfig,
    SQLCommonConfig,
    StatefulUsageConfigMixin,
    StatefulLineageConfigMixin,
    StatefulTimeWindowConfigMixin,
    StatefulProfilingConfigMixin,
    ClassificationSourceConfigMixin,
):
    include_schema_metadata: bool = Field(
        default=True,
        description="Whether to ingest the BigQuery schema, i.e. projects, schemas, tables, and views.",
    )

    usage: BigQueryUsageConfig = Field(
        default=BigQueryUsageConfig(), description="Usage related configs"
    )

    include_usage_statistics: bool = Field(
        default=True,
        description="Generate usage statistic",
    )

    capture_table_label_as_tag: Union[bool, AllowDenyPattern] = Field(
        default=False,
        description="Capture BigQuery table labels as DataHub tag",
    )

    capture_view_label_as_tag: Union[bool, AllowDenyPattern] = Field(
        default=False,
        description="Capture BigQuery view labels as DataHub tag",
    )

    capture_dataset_label_as_tag: Union[bool, AllowDenyPattern] = Field(
        default=False,
        description="Capture BigQuery dataset labels as DataHub tag",
    )

    include_table_constraints: bool = Field(
        default=True,
        description="Whether to ingest table constraints. If you know you don't use table constraints, you can disable it to save one extra query per dataset. In general it should be enabled",
    )

    include_external_url: bool = Field(
        default=True,
        description="Whether to populate BigQuery Console url to Datasets/Tables",
    )

    include_table_snapshots: Optional[bool] = Field(
        default=True, description="Whether table snapshots should be ingested."
    )

    debug_include_full_payloads: bool = Field(
        default=False,
        description="Include full payload into events. It is only for debugging and internal use.",
    )

    number_of_datasets_process_in_batch: HiddenFromDocs[int] = Field(
        default=10000,
        description="Number of table queried in batch when getting metadata. This is a low level config property "
        "which should be touched with care.",
    )

    number_of_datasets_process_in_batch_if_profiling_enabled: int = Field(
        default=1000,
        description="Number of partitioned table queried in batch when getting metadata. This is a low level config "
        "property which should be touched with care. This restriction is needed because we query "
        "partitions system view which throws error if we try to touch too many tables.",
    )

    use_tables_list_query_v2: bool = Field(
        default=False,
        description="List tables using an improved query that extracts partitions and last modified timestamps more "
        "accurately. Requires the ability to read table data. Automatically enabled when profiling is "
        "enabled.",
    )

    use_queries_v2: bool = Field(
        default=True,
        description="If enabled, uses the new queries extractor to extract queries from bigquery.",
    )
    include_queries: bool = Field(
        default=True,
        description="If enabled, generate query entities associated with lineage edges. Only applicable if `use_queries_v2` is enabled.",
    )
    include_query_usage_statistics: bool = Field(
        default=True,
        description="If enabled, generate query popularity statistics. Only applicable if `use_queries_v2` is enabled.",
    )

    @property
    def have_table_data_read_permission(self) -> bool:
        return self.use_tables_list_query_v2 or self.is_profiling_enabled()

    column_limit: int = Field(
        default=300,
        description="Maximum number of columns to process in a table. This is a low level config property which "
        "should be touched with care. This restriction is needed because excessively wide tables can "
        "result in failure to ingest the schema.",
    )

    lineage_use_sql_parser: bool = Field(
        default=True,
        description="Use sql parser to resolve view/table lineage.",
    )

    lineage_sql_parser_use_raw_names: bool = Field(
        default=False,
        description="This parameter ignores the lowercase pattern stipulated in the SQLParser. NOTE: Ignored if lineage_use_sql_parser is False.",
    )

    extract_column_lineage: bool = Field(
        default=False,
        description="If enabled, generate column level lineage. "
        "Requires lineage_use_sql_parser to be enabled.",
    )

    extract_lineage_from_catalog: bool = Field(
        default=False,
        description="This flag enables the data lineage extraction from Data Lineage API exposed by Google Data Catalog. NOTE: This extractor can't build views lineage. It's recommended to enable the view's DDL parsing. Read the docs to have more information about: https://cloud.google.com/data-catalog/docs/concepts/about-data-lineage",
    )

    extract_policy_tags_from_catalog: bool = Field(
        default=False,
        description=(
            "Extract policy tags from BigQuery tables using INFORMATION_SCHEMA and Data Catalog API. "
            "Policy tag display names are resolved by listing all tags per taxonomy (one API call per "
            "unique taxonomy, typically 3-10 calls per ingestion run). "
            "If the Data Catalog API is blocked by VPC Service Controls, policy tag resource names will "
            "be stored instead of display names (graceful degradation). "
            "Requires BigQuery API v2 (available since ~2020) and Data Catalog API access. "
            "For more information about policy tags and column-level security, refer to: "
            "https://cloud.google.com/bigquery/docs/column-level-security-intro"
        ),
    )

    scheme: str = "bigquery"

    log_page_size: PositiveInt = Field(
        default=1000,
        description="The number of log item will be queried per page for lineage collection",
    )

    # extra_client_options, include_table_lineage and max_query_duration are relevant only when computing the lineage.
    include_table_lineage: Optional[bool] = Field(
        default=True,
        description="Option to enable/disable lineage generation. Is enabled by default.",
    )

    include_column_lineage_with_gcs: bool = Field(
        default=True,
        description="When enabled, column-level lineage will be extracted from the gcs.",
    )

    max_query_duration: timedelta = Field(
        default=timedelta(minutes=15),
        description="Correction to pad start_time and end_time with. For handling the case where the read happens within our time range but the query completion event is delayed and happens after the configured end time.",
    )

    bigquery_audit_metadata_datasets: Optional[List[str]] = Field(
        default=None,
        description="A list of datasets that contain a table named cloudaudit_googleapis_com_data_access which contain BigQuery audit logs, specifically, those containing BigQueryAuditMetadata. It is recommended that the project of the dataset is also specified, for example, projectA.datasetB.",
    )
    use_exported_bigquery_audit_metadata: bool = Field(
        default=False,
        description="When configured, use BigQueryAuditMetadata in bigquery_audit_metadata_datasets to compute lineage information.",
    )
    use_date_sharded_audit_log_tables: bool = Field(
        default=False,
        description="Whether to read date sharded tables or time partitioned tables when extracting usage from exported audit logs.",
    )

    _cache_path: Optional[str] = PrivateAttr(None)

    upstream_lineage_in_report: bool = Field(
        default=False,
        description="Useful for debugging lineage information. Set to True to see the raw lineage created internally. Only works with legacy approach (`use_queries_v2: False`).",
    )

    run_optimized_column_query: HiddenFromDocs[bool] = Field(
        default=False,
        description="Run optimized column query to get column information. This is an experimental feature and may not work for all cases.",
    )

    file_backed_cache_size: HiddenFromDocs[int] = Field(
        default=2000,
        description="Maximum number of entries for the in-memory caches of FileBacked data structures.",
    )

    convert_column_urns_to_lowercase: bool = Field(
        default=False,
        description="When enabled, converts column URNs to lowercase to ensure cross-platform compatibility.",
    )

    exclude_empty_projects: bool = Field(
        default=False,
        description="Option to exclude empty projects from being ingested.",
    )

    schema_resolution_batch_size: HiddenFromDocs[int] = Field(
        default=100,
        description="The number of tables to process in a batch when resolving schema from DataHub.",
    )

    max_threads_dataset_parallelism: int = Field(
        default=DEFAULT_BQ_SCHEMA_PARALLELISM,
        description="Number of worker threads to use to parallelize BigQuery Dataset Metadata Extraction."
        " Set to 1 to disable.",
    )

    region_qualifiers: List[str] = Field(
        default=["region-us", "region-eu"],
        description="BigQuery regions to be scanned for bigquery jobs when using `use_queries_v2`. "
        "See [this](https://cloud.google.com/bigquery/docs/information-schema-jobs#scope_and_syntax) for details.",
    )

    pushdown_deny_usernames: List[str] = Field(
        default=[],
        description="List of user email patterns using SQL LIKE syntax (e.g., 'bot_%', '%@%.iam.gserviceaccount.com') "
        "which will NOT be considered for lineage/usage/queries extraction. "
        "Uses case-insensitive LIKE for server-side filtering (e.g., 'bot_%' matches 'Bot_User@example.com'). "
        "This is primarily useful for improving performance by filtering out users with extremely high query volumes. "
        "Only applicable if `use_queries_v2` is enabled.",
    )

    pushdown_allow_usernames: List[str] = Field(
        default=[],
        description="List of user email patterns using SQL LIKE syntax (e.g., 'analyst_%@company.com') "
        "which WILL be considered for lineage/usage/queries extraction. "
        "Uses case-insensitive LIKE for server-side filtering (e.g., 'analyst_%' matches 'Analyst_John@company.com'). "
        "Only applicable if `use_queries_v2` is enabled. If not specified, all users not in deny list are included.",
    )

    _include_view_lineage = pydantic_removed_field(
        "include_view_lineage", month="January", year=2025
    )
    _include_view_column_lineage = pydantic_removed_field(
        "include_view_column_lineage", month="January", year=2025
    )
    _lineage_parse_view_ddl = pydantic_removed_field(
        "lineage_parse_view_ddl", month="January", year=2025
    )

    @model_validator(mode="before")
    @classmethod
    def set_include_schema_metadata(cls, values: Dict) -> Dict:
        # Create a copy to avoid modifying the input dictionary, preventing state contamination in tests
        values = deepcopy(values)
        # Historically this is used to disable schema ingestion
        if (
            "include_tables" in values
            and "include_views" in values
            and not values["include_tables"]
            and not values["include_views"]
        ):
            values["include_schema_metadata"] = False
            values["include_table_snapshots"] = False
            logger.info(
                "include_tables and include_views are both set to False."
                " Disabling schema metadata ingestion for tables, views, and snapshots."
            )

        return values

    @model_validator(mode="before")
    @classmethod
    def profile_default_settings(cls, values: Dict) -> Dict:
        # Create a copy to avoid modifying the input dictionary, preventing state contamination in tests
        values = deepcopy(values)
        # Extra default SQLAlchemy option for better connection pooling and threading.
        # https://docs.sqlalchemy.org/en/14/core/pooling.html#sqlalchemy.pool.QueuePool.params.max_overflow
        values.setdefault("options", {}).setdefault("max_overflow", -1)

        return values

    @field_validator("bigquery_audit_metadata_datasets", mode="after")
    @classmethod
    def validate_bigquery_audit_metadata_datasets(
        cls, v: Optional[List[str]], info: ValidationInfo
    ) -> Optional[List[str]]:
        if info.data.get("use_exported_bigquery_audit_metadata"):
            assert v and len(v) > 0, (
                "`bigquery_audit_metadata_datasets` should be set if using `use_exported_bigquery_audit_metadata: True`."
            )

        return v

    @field_validator("upstream_lineage_in_report", mode="after")
    @classmethod
    def validate_upstream_lineage_in_report(cls, v: bool, info: ValidationInfo) -> bool:
        if v and info.data.get("use_queries_v2", True):
            logging.warning(
                "`upstream_lineage_in_report` is enabled but will be ignored because `use_queries_v2` is enabled."
                "This debugging feature only works with the legacy lineage approach (`use_queries_v2: false`)."
            )

        return v

    @model_validator(mode="after")
    def validate_queries_v2_stateful_ingestion(self) -> "BigQueryV2Config":
        if self.use_queries_v2:
            if (
                self.enable_stateful_lineage_ingestion
                or self.enable_stateful_usage_ingestion
            ):
                logger.warning(
                    "enable_stateful_lineage_ingestion and enable_stateful_usage_ingestion are deprecated "
                    "when using use_queries_v2=True. These configs only work with the legacy (non-queries v2) extraction path. "
                    "For queries v2, use enable_stateful_time_window instead to enable stateful ingestion "
                    "for the unified time window extraction (lineage + usage + operations + queries)."
                )
        return self

    @field_validator(
        "pushdown_deny_usernames", "pushdown_allow_usernames", mode="after"
    )
    @classmethod
    def validate_pushdown_username_patterns(cls, patterns: List[str]) -> List[str]:
        """Validate and normalize pushdown username patterns.

        - Strips leading/trailing whitespace from each pattern
        - Rejects empty patterns (after stripping)

        Note: Patterns use SQL LIKE syntax (% = any characters, _ = single character).
        Invalid patterns will fail at runtime when the SQL query is executed.
        """
        validated = []
        for i, pattern in enumerate(patterns):
            stripped = pattern.strip()
            if not stripped:
                raise ValueError(
                    f"Empty pattern at index {i}. "
                    "Remove empty strings from pushdown username patterns."
                )
            validated.append(stripped)
        return validated

    @model_validator(mode="after")
    def validate_pushdown_config(self) -> "BigQueryV2Config":
        if (
            self.pushdown_deny_usernames or self.pushdown_allow_usernames
        ) and not self.use_queries_v2:
            raise ValueError(
                "pushdown_deny_usernames and pushdown_allow_usernames require use_queries_v2=True. "
                "Either enable use_queries_v2 or remove the pushdown username filters."
            )
        return self

    def get_table_pattern(self, pattern: List[str]) -> str:
        return "|".join(pattern) if pattern else ""

    _platform_instance_not_supported_for_bigquery = pydantic_removed_field(
        "platform_instance",
        month="September",
        year=2025,
    )



"""Data pack models for the DataHub CLI."""

from typing import List, Optional

from pydantic import ConfigDict, Field

from datahub.configuration.common import ConfigModel
from datahub.utilities.str_enum import StrEnum


class TrustTier(StrEnum):
    """Trust level of a data pack."""

    VERIFIED = "verified"
    COMMUNITY = "community"
    CUSTOM = "custom"


class DataPackInfo(ConfigModel):
    """Metadata about a data pack available for loading into DataHub."""

    name: str = Field(
        description="Pack identifier (e.g. 'bootstrap', 'showcase-ecommerce')"
    )
    description: str = Field(
        description="Human-readable description of the pack contents"
    )
    url: str = Field(description="HTTP(S) URL to the MCP/MCE JSON data file")
    sha256: Optional[str] = Field(
        default=None,
        description="Hex-encoded SHA256 checksum of the data file for integrity verification",
    )
    size_hint: Optional[str] = Field(
        default=None,
        description="Approximate download size for display (e.g. '3.2 MB')",
    )
    tags: List[str] = Field(default_factory=list, description="Categorization tags")
    source_type: str = Field(
        default="http",
        description="Source type: 'http', 's3', or 'huggingface' (all resolve to HTTP in v1)",
    )
    trust: TrustTier = Field(
        default=TrustTier.CUSTOM,
        description="Trust tier: verified (DataHub official), community, or custom",
    )
    reference_timestamp: Optional[int] = Field(
        default=None,
        description="Epoch millis when the pack was captured. Used as the anchor for time-shifting.",
    )
    min_server_version: Optional[str] = Field(
        default=None,
        description="Minimum DataHub OSS server version required (e.g. '0.14.0')",
    )
    min_cloud_version: Optional[str] = Field(
        default=None,
        description="Minimum Acryl Cloud version required (e.g. '0.3.5')",
    )
    pack_format_version: str = Field(
        default="1", description="Schema version of the pack manifest"
    )


class RegistryManifest(ConfigModel):
    """Top-level structure of the remote registry JSON file."""

    model_config = ConfigDict(extra="allow")

    schema_version: int = Field(
        default=1,
        description="Registry schema version. Old clients ignore unknown fields.",
    )
    packs: dict[str, DataPackInfo] = Field(
        default_factory=dict, description="Map of pack name to pack info"
    )


class LoadRecord(ConfigModel):
    """Tracks a data pack load for unload support."""

    pack_name: str = Field(description="Name of the loaded pack")
    run_id: str = Field(description="Ingestion run ID used during load")
    loaded_at: str = Field(description="ISO 8601 timestamp of when the pack was loaded")
    pack_url: str = Field(description="URL the pack was downloaded from")
    pack_sha256: Optional[str] = Field(
        default=None, description="SHA256 of the data file that was loaded"
    )



from dataclasses import dataclass, field
from typing import Dict, List, Optional

import pydantic
from pydantic import BaseModel, Field

from datahub.configuration.common import AllowDenyPattern, TransparentSecretStr
from datahub.configuration.source_common import (
    EnvConfigMixin,
    PlatformInstanceConfigMixin,
)
from datahub.ingestion.api.report import EntityFilterReport, Report
from datahub.ingestion.source.state.stale_entity_removal_handler import (
    StaleEntityRemovalSourceReport,
    StatefulStaleMetadataRemovalConfig,
)
from datahub.ingestion.source.state.stateful_ingestion_base import (
    StatefulIngestionConfigBase,
)


class Constant:
    """
    keys used in sigma plugin
    """

    # Rest API response key constants
    REFRESH_TOKEN = "refresh_token"
    ACCESS_TOKEN = "access_token"
    EMAIL = "email"
    ENTRIES = "entries"
    MEMBERID = "memberId"
    EDGES = "edges"
    DEPENDENCIES = "dependencies"
    SOURCE = "source"
    TARGET = "target"
    WORKSPACEID = "workspaceId"
    PATH = "path"
    NAME = "name"
    URL = "url"
    ELEMENTID = "elementId"
    ID = "id"
    PARENTID = "parentId"
    TYPE = "type"
    DATASET = "dataset"
    WORKBOOK = "workbook"
    BADGE = "badge"
    NEXTPAGE = "nextPage"
    NEXTPAGETOKEN = "nextPageToken"
    DATA_MODEL = "data-model"
    DATA_MODEL_ID = "dataModelId"

    # Source Config constants
    DEFAULT_API_URL = "https://aws-api.sigmacomputing.com/v2"


class WorkspaceCounts(BaseModel):
    workbooks_count: int = 0
    datasets_count: int = 0
    elements_count: int = 0
    pages_count: int = 0
    data_models_count: int = 0
    data_model_elements_count: int = 0

    def is_empty(self) -> bool:
        return (
            self.workbooks_count == 0
            and self.datasets_count == 0
            and self.elements_count == 0
            and self.pages_count == 0
            and self.data_models_count == 0
            and self.data_model_elements_count == 0
        )

    def as_obj(self) -> dict:
        return {
            "workbooks_count": self.workbooks_count,
            "datasets_count": self.datasets_count,
            "elements_count": self.elements_count,
            "pages_count": self.pages_count,
            "data_models_count": self.data_models_count,
            "data_model_elements_count": self.data_model_elements_count,
        }


class SigmaWorkspaceEntityFilterReport(EntityFilterReport):
    type: str = "workspace"

    workspace_counts: Dict[str, WorkspaceCounts] = Field(
        default_factory=dict,
        description="Counts of workbooks, datasets, elements and pages in each workspace.",
    )

    def increment_workbooks_count(self, workspace_id: str) -> None:
        if workspace_id not in self.workspace_counts:
            self.workspace_counts[workspace_id] = WorkspaceCounts()
        self.workspace_counts[workspace_id].workbooks_count += 1

    def increment_datasets_count(self, workspace_id: str) -> None:
        if workspace_id not in self.workspace_counts:
            self.workspace_counts[workspace_id] = WorkspaceCounts()
        self.workspace_counts[workspace_id].datasets_count += 1

    def increment_elements_count(self, workspace_id: str) -> None:
        if workspace_id not in self.workspace_counts:
            self.workspace_counts[workspace_id] = WorkspaceCounts()
        self.workspace_counts[workspace_id].elements_count += 1

    def increment_pages_count(self, workspace_id: str) -> None:
        if workspace_id not in self.workspace_counts:
            self.workspace_counts[workspace_id] = WorkspaceCounts()
        self.workspace_counts[workspace_id].pages_count += 1

    def increment_data_models_count(self, workspace_id: str) -> None:
        if workspace_id not in self.workspace_counts:
            self.workspace_counts[workspace_id] = WorkspaceCounts()
        self.workspace_counts[workspace_id].data_models_count += 1

    def increment_data_model_elements_count(self, workspace_id: str) -> None:
        if workspace_id not in self.workspace_counts:
            self.workspace_counts[workspace_id] = WorkspaceCounts()
        self.workspace_counts[workspace_id].data_model_elements_count += 1

    def as_obj(self) -> dict:
        return {
            "filtered": self.dropped_entities.as_obj(),
            "processed": self.processed_entities.as_obj(),
            "workspace_counts": {
                key: item.as_obj() for key, item in self.workspace_counts.items()
            },
        }


@dataclass
class ElementDmEdgeReport(Report):
    """Workbook-to-DM-element bridge counters. Matched by element name."""

    # Edge emitted as a unique chart input.
    resolved: int = 0
    # URN already present on the chart (e.g. diamond lineage).
    deduped: int = 0
    # Multiple elements share the name; deterministic pick.
    ambiguous: int = 0
    # DM ingested but name did not match (no edge; ChartInfo.inputs requires
    # Dataset URNs, not Container URNs).
    name_unmatched_but_dm_known: int = 0
    # Lineage node had no ``name``; distinct from the rename-miss counter.
    upstream_name_missing: int = 0
    # DM not in this run.
    unresolved: int = 0
    # Sigma places the DM-reference node only in ``edges[].source`` (not as a
    # ``dependencies`` key). We synthesize the upstream using the workbook
    # element's own ``name`` (Sigma's default mirrors the DM element name).
    synthesized_from_edge_only: int = 0


@dataclass
class SigmaSourceReport(StaleEntityRemovalSourceReport):
    workspaces: SigmaWorkspaceEntityFilterReport = field(
        default_factory=SigmaWorkspaceEntityFilterReport
    )
    non_accessible_workspaces_count: int = 0

    datasets: EntityFilterReport = EntityFilterReport.field(type="dataset")
    datasets_without_workspace: int = 0

    workbooks: EntityFilterReport = EntityFilterReport.field(type="workbook")
    workbooks_without_workspace: int = 0

    data_models: EntityFilterReport = EntityFilterReport.field(type="data_model")
    data_models_without_workspace: int = 0

    number_of_files_metadata: Dict[str, int] = field(default_factory=dict)
    empty_workspaces: List[str] = field(default_factory=list)

    # Sheet upstream skipped because the upstream element was filtered out
    # of the chart map (e.g. pivot-table blocked by page-element allowlist).
    num_filtered_sheet_upstreams: int = 0

    # Chart upstream node of type=dataset had ``name=None``. The SQL-bridge
    # cannot correlate it against a warehouse table, so the chart-to-Sigma-
    # dataset edge is skipped. Pre-PR this raised ValidationError; this
    # counter restores the observability signal.
    chart_dataset_upstream_name_missing: int = 0

    # Chart InputFields — one counter fires per chart column (not per formula ref).
    # The resolver (_resolve_chart_formula_upstream) is a pure predicate: it
    # returns a resolved (urn, field) pair or None; all counting happens in
    # _build_element_input_fields so every column lands in exactly one bucket.
    # Invariant: resolved + self_ref_fallback + skipped_parameter + skipped_sibling
    #            == total chart columns processed across the workbook.
    chart_input_fields_resolved: int = 0
    # Column emitted with self-referential schemaFieldUrn because no formula ref
    # resolved (includes: no-formula column, unresolvable ref, mixed param+real
    # where the real refs fail). Keeps V2 column list visible unconditionally.
    chart_input_fields_self_ref_fallback: int = 0
    # Column whose formula refs are exclusively parameter refs (e.g. [P_*]).
    chart_input_fields_skipped_parameter: int = 0
    # Column whose formula refs are exclusively bare sibling refs (e.g. [col]).
    chart_input_fields_skipped_sibling: int = 0
    # Sub-bucket of self_ref_fallback: source name that is a case-only mismatch
    # against a workbook element name (warehouse fallback intentionally skipped).
    chart_input_fields_case_mismatch: int = 0
    # Workbooks whose /columns pagination aborted partway through. InputFields
    # for those workbooks may be missing columns that appear after the failure.
    column_formulas_fetch_partial: int = 0

    # DM element emission / upstream resolution.
    data_model_elements_emitted: int = 0
    data_model_element_intra_upstreams: int = 0
    data_model_element_external_upstreams: int = 0
    # Split intentionally so operators can triage "upstream dataset
    # exists but wasn't ingested in this run" (typically a pattern
    # filter or missing read perm) vs "source_id shape we do not yet
    # parse" (cross-DM refs ahead of the follow-up PR, or a future
    # Sigma shape). ``data_model_element_upstreams_unresolved`` is
    # kept as an aggregate for dashboards that already read it.
    data_model_element_upstreams_unresolved_external: int = 0
    data_model_element_upstreams_unknown_shape: int = 0
    data_model_element_upstreams_unresolved: int = 0

    # Cross-DM element references (DM-A element pulls from DM-B). Success
    # counters bump once per unique upstream URN (diamonds deduped).
    # ``_resolved = _strict + _ambiguous + _single_element_fallback``
    # (``_ambiguous`` and ``_single_element_fallback`` are sub-shapes of
    # ``_resolved``; clean strict-name-match = _resolved - _ambiguous -
    # _single_element_fallback).
    # Failure counters bump per source_id (each is a distinct missing ref)
    # and also bump ``data_model_element_upstreams_unresolved``:
    # ``_name_unmatched_but_dm_known``: producer DM ingested, no name match
    #   (typically a consumer-side rename).
    # ``_dm_unknown``: producer DM not in this run.
    # ``_consumer_name_missing``: consumer element had a blank name; the
    #   name-bridge was never attempted.
    # ``_self_reference``: producer prefix matches the consuming DM (API
    #   payload anomaly; defensively skipped).
    data_model_element_cross_dm_upstreams_resolved: int = 0
    data_model_element_cross_dm_upstreams_ambiguous: int = 0
    data_model_element_cross_dm_upstreams_single_element_fallback: int = 0
    data_model_element_cross_dm_upstreams_name_unmatched_but_dm_known: int = 0
    data_model_element_cross_dm_upstreams_dm_unknown: int = 0
    data_model_element_cross_dm_upstreams_consumer_name_missing: int = 0
    data_model_element_cross_dm_upstreams_self_reference: int = 0

    # Personal-space / unlisted DMs discovered via /v2/dataModels/{urlId}.
    # ``_discovered``: an unlisted DM was fetched and added to the run.
    # ``_unresolved``: fetch returned non-200 (usually 403 / 404).
    # ``_rate_limited``: fetch returned 429 after the urllib3 retry budget
    #   was exhausted -- sub-count of ``_unresolved`` surfaced separately
    #   so operators can distinguish "Sigma is rate-limiting us" (transient,
    #   re-run the job) from "the DM is genuinely forbidden or deleted"
    #   (steady-state). ``_unresolved`` is still bumped for 429 so the
    #   aggregate stays accurate.
    data_model_external_references_discovered: int = 0
    data_model_external_reference_unresolved: int = 0
    data_model_external_reference_rate_limited: int = 0

    # /columns entries with ``elementId = None`` (DM-global calculations),
    # dropped because there is no element Dataset to attach them to.
    data_model_columns_without_element_dropped: int = 0

    # Two DMs claimed the same ``urlId`` bridge key. The first wins; the
    # second is skipped at emit time to avoid an unlinked orphan. Non-zero
    # means a reissued slug; see warning log for ``(dataModelId, urlId)``.
    data_models_bridge_key_collision: int = 0

    # Duplicate ``column.name`` on a single DM element, dropped to avoid
    # ``SchemaMetadata`` with duplicate ``fieldPath`` values.
    data_model_element_columns_duplicate_fieldpath_dropped: int = 0

    # DM element column-level lineage (FGL) counters.
    # How many FGL entries were emitted across all elements.
    data_model_element_fgl_emitted: int = 0
    # Refs where multiple sibling candidates passed the /lineage filter;
    # sorted-first URN was chosen (matches T2 PR1's collision precedent).
    data_model_element_fgl_collision_pick_first: int = 0
    # Refs whose source element is outside this DM; deferred to cross-DM resolution.
    data_model_element_fgl_cross_dm_deferred: int = 0
    # Refs where element-name matches the element's own warehouse-table name
    # (e.g., element "data.csv" with formula "[data.csv/col]"). These are
    # warehouse-passthrough passthroughs, not intra-DM self-edges; the actual
    # upstream is the warehouse inode — out of RESOLVE-A scope.
    data_model_element_fgl_warehouse_passthrough_deferred: int = 0
    # Refs whose source element is in this DM but not listed as an upstream by
    # /lineage; dropped to avoid orphan FGL the UI silently rejects.
    data_model_element_fgl_dropped_orphan_upstream: int = 0
    # Refs whose column name has no matching fieldPath in the upstream element's
    # schema; dropped to avoid a dangling schemaField URN.
    data_model_element_fgl_dropped_unknown_upstream_column: int = 0

    # Entries dropped as duplicates by the pagination-level natural-key
    # dedup in ``_paginated_entries`` / lineage raw dedup. Normally 0;
    # non-zero indicates an echoed pagination cursor or server-side
    # overlap between pages -- correctness is preserved (no double
    # emission) but the signal is surfaced here so operators can spot it.
    pagination_duplicate_entries_dropped: int = 0
    # Entries dropped by per-endpoint ``ValidationError`` handling. Only
    # the first ``_MAX_MALFORMED_WARNINGS_PER_ENDPOINT`` rows per endpoint
    # emit a user-visible warning to prevent report flooding on a
    # vendor-wide regression; this counter captures the rest.
    pagination_malformed_entries_dropped: int = 0

    element_dm_edge: ElementDmEdgeReport = field(default_factory=ElementDmEdgeReport)


class PlatformDetail(PlatformInstanceConfigMixin, EnvConfigMixin):
    data_source_platform: str = pydantic.Field(
        description="A chart's data sources platform name.",
    )
    default_db: Optional[str] = pydantic.Field(
        default=None,
        description="Default database name to use when parsing SQL queries. "
        "Used to generate fully qualified table URNs (e.g., 'prod' for 'prod.public.table').",
    )
    default_schema: Optional[str] = pydantic.Field(
        default=None,
        description="Default schema name to use when parsing SQL queries. "
        "Used to generate fully qualified table URNs (e.g., 'public' for 'prod.public.table').",
    )


class SigmaSourceConfig(
    StatefulIngestionConfigBase, PlatformInstanceConfigMixin, EnvConfigMixin
):
    api_url: str = pydantic.Field(
        default=Constant.DEFAULT_API_URL, description="Sigma API hosted URL."
    )
    client_id: str = pydantic.Field(description="Sigma Client ID")
    client_secret: TransparentSecretStr = pydantic.Field(
        description="Sigma Client Secret"
    )
    # Sigma workspace identifier
    workspace_pattern: AllowDenyPattern = pydantic.Field(
        default=AllowDenyPattern.allow_all(),
        description="Regex patterns to filter Sigma workspaces in ingestion."
        "Mention 'My documents' if personal entities also need to ingest.",
    )
    ingest_owner: Optional[bool] = pydantic.Field(
        default=True,
        description="Ingest Owner from source. This will override Owner info entered from UI.",
    )
    ingest_shared_entities: Optional[bool] = pydantic.Field(
        default=False,
        description="Whether to ingest the shared entities or not.",
    )
    extract_lineage: Optional[bool] = pydantic.Field(
        default=True,
        description="Whether to extract lineage of workbook's elements and datasets or not.",
    )
    workbook_lineage_pattern: AllowDenyPattern = pydantic.Field(
        default=AllowDenyPattern.allow_all(),
        description="Regex patterns to filter workbook's elements and datasets lineage in ingestion."
        "Requires extract_lineage to be enabled.",
    )
    chart_sources_platform_mapping: Dict[str, PlatformDetail] = pydantic.Field(
        default={},
        description="A mapping of the sigma workspace/workbook/chart folder path to all chart's data sources platform details present inside that folder path.",
    )
    stateful_ingestion: Optional[StatefulStaleMetadataRemovalConfig] = pydantic.Field(
        default=None, description="Sigma Stateful Ingestion Config."
    )
    workbook_pattern: AllowDenyPattern = pydantic.Field(
        default=AllowDenyPattern.allow_all(),
        description="Regex patterns to filter Sigma workbook names in ingestion.",
    )
    ingest_data_models: bool = pydantic.Field(
        default=False,
        description="Whether to ingest Sigma Data Models. Each Data Model is emitted "
        "as a Container with one Dataset per element inside it (plus per-element "
        "``SchemaMetadata`` and, when ``extract_lineage`` is also enabled, "
        "``UpstreamLineage``). Default is ``False`` because "
        "enabling this introduces a new entity class to the graph — existing tenants "
        "will see new Containers and Datasets appear on first ingest and will need "
        "to factor those into any soft-delete policy if they later disable this flag. "
        "Enabling this issues ``/dataModels/{id}/elements`` and ``/columns`` calls "
        "per Data Model unconditionally; the ``/lineage`` call is only issued when "
        "``extract_lineage`` is also ``True`` (so users who opt out of lineage at "
        "the workbook surface don't get a lineage endpoint hit under a different flag).",
    )
    data_model_pattern: AllowDenyPattern = pydantic.Field(
        default=AllowDenyPattern.allow_all(),
        description="Regex patterns to filter Sigma Data Model names in ingestion. "
        "Requires ingest_data_models to be enabled.",
    )
    max_personal_dm_discovery_rounds: int = pydantic.Field(
        default=20,
        ge=1,
        description="Belt-and-braces safety cap on the number of passes the "
        "personal-space Data Model discovery loop is allowed to make. Each "
        "pass fetches ``/v2/dataModels/{urlId}`` for every newly-seen "
        "cross-DM ``<urlId>`` prefix; the loop terminates naturally when "
        "``unresolved_seen`` plateaus (monotonically growing set), so under "
        "a well-behaved API this cap is never hit. Must be ``>= 1`` -- set "
        "``ingest_shared_entities: False`` (or leave it at the default) if "
        "the goal is to disable personal-space discovery entirely; ``0`` / "
        "negative values are rejected because the first pass is required "
        "to prepopulate the bridge maps for listed DMs. Exists to protect "
        "against pathological Sigma payloads (e.g. a chain of personal-space "
        "DMs that keep referencing newly-discovered personal-space DMs) by "
        "breaking with a ``SourceReport.warning`` instead of looping "
        "unbounded.",
    )



import logging
from dataclasses import dataclass
from typing import TYPE_CHECKING, Optional

from sqlalchemy.exc import DatabaseError, OperationalError, ProgrammingError

from datahub.ingestion.source.sql.postgres.query import PostgresQuery
from datahub.metadata.urns import CorpUserUrn
from datahub.sql_parsing.sql_parsing_aggregator import (
    ObservedQuery,
    SqlParsingAggregator,
)
from datahub.sql_parsing.sqlglot_lineage import (
    SqlUnderstandingError,
    UnsupportedStatementTypeError,
)
from datahub.utilities.perf_timer import PerfTimer

if TYPE_CHECKING:
    from sqlalchemy.engine import Connection

    from datahub.ingestion.source.sql.postgres.source import PostgresConfig
    from datahub.ingestion.source.sql.sql_common import SQLSourceReport

logger = logging.getLogger(__name__)


@dataclass
class PostgresQueryEntry:
    """Represents a single query entry from pg_stat_statements."""

    query_id: str
    query_text: str
    execution_count: int
    total_exec_time_ms: float
    user_name: Optional[str]
    database_name: str

    @property
    def avg_exec_time_ms(self) -> float:
        """Calculate average execution time."""
        return (
            self.total_exec_time_ms / self.execution_count
            if self.execution_count > 0
            else 0.0
        )


class PostgresLineageExtractor:
    """
    Extracts lineage from Postgres query history.

    Uses pg_stat_statements to capture executed queries and feeds them
    to SqlParsingAggregator for lineage extraction via sqlglot.
    """

    def __init__(
        self,
        config: "PostgresConfig",
        connection: "Connection",
        report: "SQLSourceReport",
        sql_aggregator: SqlParsingAggregator,
        default_schema: str = "public",
    ) -> None:
        self.config = config
        self.connection = connection
        self.report = report
        self.sql_aggregator = sql_aggregator
        self.default_schema = default_schema

        self.queries_extracted = 0
        self.queries_parsed = 0
        self.queries_failed = 0

    def check_prerequisites(self) -> tuple[bool, str]:
        """Verify pg_stat_statements is properly configured."""
        try:
            result = self.connection.execute(PostgresQuery.get_postgres_version())
            row = result.fetchone()

            if row:
                version_num = row[0]
                if version_num < 130000:
                    return (
                        False,
                        f"PostgreSQL version {version_num // 10000}.{(version_num // 100) % 100} detected. "
                        "Query-based lineage requires PostgreSQL 13+ due to column name changes in pg_stat_statements "
                        "(total_time -> total_exec_time). Please upgrade to PostgreSQL 13 or later.",
                    )

        except (DatabaseError, OperationalError) as e:
            logger.warning("Failed to check PostgreSQL version: %s", e)

        try:
            result = self.connection.execute(
                PostgresQuery.check_pg_stat_statements_enabled()
            )
            row = result.fetchone()

            if not row or not row[0]:
                return (
                    False,
                    "pg_stat_statements extension not installed. "
                    "Install with: CREATE EXTENSION pg_stat_statements;",
                )

        except (DatabaseError, OperationalError) as e:
            logger.warning("Failed to check pg_stat_statements extension: %s", e)
            return False, f"Extension check failed: {e}"

        try:
            result = self.connection.execute(
                PostgresQuery.check_pg_stat_statements_permissions()
            )
            row = result.fetchone()

            if row:
                has_stats_role = row[0]
                is_superuser = row[1]

                if not (has_stats_role or is_superuser):
                    return (
                        False,
                        "Insufficient permissions. Grant pg_read_all_stats role: "
                        "GRANT pg_read_all_stats TO <user>;",
                    )
            else:
                return (
                    False,
                    "Could not verify pg_stat_statements permissions. "
                    "Unable to determine user roles.",
                )

        except (DatabaseError, OperationalError) as e:
            logger.error("Failed to check permissions: %s", e)
            return False, f"Permission check failed: {e}"

        return True, "Prerequisites met"

    def extract_query_history(self) -> list[PostgresQueryEntry]:
        """Extract queries from pg_stat_statements ordered by execution time/count."""
        is_ready, message = self.check_prerequisites()
        if not is_ready:
            logger.error(
                "pg_stat_statements not ready: %s. Query-based lineage will be skipped.",
                message,
            )
            self.report.report_failure(
                message=message,
                context="pg_stat_statements_not_ready",
            )
            return []

        logger.info("Prerequisites check: %s", message)

        query, params = PostgresQuery.get_query_history(
            database=self.config.database,
            limit=self.config.max_queries_to_extract,
            min_calls=self.config.min_query_calls,
            exclude_patterns=self.config.query_exclude_patterns,
        )

        with PerfTimer() as timer:
            try:
                result = self.connection.execute(query, params)

                queries = []
                for row in result:
                    self.queries_extracted += 1

                    queries.append(
                        PostgresQueryEntry(
                            query_id=row["query_id"],
                            query_text=row["query_text"],
                            execution_count=row["execution_count"],
                            total_exec_time_ms=row["total_exec_time_ms"],
                            user_name=row["user_name"],
                            database_name=row["database_name"],
                        )
                    )

                logger.info(
                    "Extracted %d queries from pg_stat_statements in %.2f seconds",
                    self.queries_extracted,
                    timer.elapsed_seconds(),
                )

                self.report.num_queries_extracted = self.queries_extracted
                return queries

            except (DatabaseError, OperationalError, ProgrammingError) as e:
                # Expected database errors: connection issues, permission denied, invalid SQL, etc.
                logger.error("Failed to extract query history: %s", e)
                self.report.report_failure(
                    message=str(e),
                    context="query_history_extraction_failed",
                )
                return []

    def populate_lineage_from_queries(self) -> None:
        """Extract lineage from query history and add to SQL aggregator."""
        if not self.config.include_query_lineage:
            logger.info("Query-based lineage extraction disabled in config")
            return

        logger.info(
            "Starting query-based lineage extraction (max_queries=%s)",
            self.config.max_queries_to_extract,
        )

        queries = self.extract_query_history()

        with PerfTimer() as timer:
            for query_entry in queries:
                try:
                    self.sql_aggregator.add_observed_query(
                        ObservedQuery(
                            query=query_entry.query_text,
                            default_db=query_entry.database_name,
                            default_schema=self.default_schema,
                            timestamp=None,
                            user=(
                                CorpUserUrn(query_entry.user_name)
                                if query_entry.user_name
                                else None
                            ),
                            session_id=f"queryid:{query_entry.query_id}",
                        )
                    )

                    self.queries_parsed += 1

                except (
                    SqlUnderstandingError,
                    UnsupportedStatementTypeError,
                ) as e:
                    logger.warning(
                        "Failed to parse query %s: %s. Query: %s...",
                        query_entry.query_id,
                        e,
                        query_entry.query_text[:100],
                    )
                    self.queries_failed += 1
                except (ValueError, KeyError) as e:
                    logger.error(
                        "Unexpected error processing query %s: %s (%s). Query: %s... "
                        "This may indicate a bug - please report this issue.",
                        query_entry.query_id,
                        e,
                        type(e).__name__,
                        query_entry.query_text[:100],
                        exc_info=True,
                    )
                    self.queries_failed += 1

        logger.info(
            "Processed %d queries for lineage extraction (%d failed) in %.2f seconds",
            self.queries_parsed,
            self.queries_failed,
            timer.elapsed_seconds(),
        )

        self.report.num_queries_parsed = self.queries_parsed
        self.report.num_queries_parse_failures = self.queries_failed



from typing import Any, Callable, Iterable, List, Optional, TypeVar

_T = TypeVar("_T")


def deduplicate_list(
    iterable: Iterable[_T],
    key: Optional[Callable[[_T], Any]] = None,
) -> List[_T]:
    """
    Remove duplicates from an iterable, preserving order.
    This serves as a replacement for OrderedSet, which is broken in Python 3.10.
    """
    seen = set()
    result: List[_T] = []
    for item in iterable:
        k = key(item) if key is not None else item
        if k not in seen:
            seen.add(k)
            result.append(item)
    return result



"""MySQL-specific profiling adapter."""

import logging
from typing import Any, Optional

import sqlalchemy as sa
from sqlalchemy.engine import Connection
from sqlalchemy.exc import SQLAlchemyError
from sqlalchemy.sql.elements import ColumnElement

from datahub.ingestion.source.sqlalchemy_profiler.base_adapter import PlatformAdapter

logger = logging.getLogger(__name__)


class MySQLAdapter(PlatformAdapter):
    """
    MySQL-specific profiling adapter.

    MySQL features:
    1. Fast row count estimation via information_schema.tables.table_rows
    2. Standard SQL for most operations
    3. No native MEDIAN function

    Uses default setup_profiling and cleanup from PlatformAdapter.
    """

    # =========================================================================
    # SQL Expression Builders
    # =========================================================================

    def get_approx_unique_count_expr(self, column: str) -> ColumnElement[Any]:
        """
        MySQL approximate unique count - uses exact COUNT(DISTINCT).

        MySQL doesn't have a built-in approximate count distinct function,
        so we fall back to exact count.

        Args:
            column: Column name

        Returns:
            SQLAlchemy expression for COUNT(DISTINCT column)
        """
        return sa.func.count(sa.func.distinct(sa.column(column)))

    def get_median_expr(self, column: str) -> Optional[ColumnElement[Any]]:
        """
        MySQL median - not supported by default.

        Standard MySQL doesn't have a MEDIAN function. It can be
        implemented using subqueries but adds complexity.

        Args:
            column: Column name

        Returns:
            None (not supported)
        """
        return None

    # =========================================================================
    # Row Count Estimation
    # =========================================================================

    def supports_row_count_estimation(self) -> bool:
        """
        MySQL supports fast row count estimation.

        Returns:
            True - MySQL has information_schema.tables.table_rows
        """
        return True

    def get_estimated_row_count(
        self, table: sa.Table, conn: Connection
    ) -> Optional[int]:
        """
        Get fast row count estimate using information_schema.tables.table_rows.

        This avoids a full table scan by using MySQL's statistics.
        The estimate may be slightly out of date but is very fast.

        Args:
            table: SQLAlchemy table object
            conn: Active database connection

        Returns:
            Estimated row count, or None if query fails
        """
        try:
            schema = table.schema
            table_name = table.name

            # Query information_schema.tables using SQLAlchemy query builder
            info_schema_tables = sa.Table(
                "tables",
                sa.MetaData(),
                sa.Column("table_schema", sa.String),
                sa.Column("table_name", sa.String),
                sa.Column("table_rows", sa.BigInteger),
                schema="information_schema",
            )
            query = (
                sa.select([info_schema_tables.c.table_rows])
                .where(info_schema_tables.c.table_schema == schema)
                .where(info_schema_tables.c.table_name == table_name)
            )

            result = conn.execute(query).scalar()
            return int(result) if result is not None else None

        except SQLAlchemyError as e:
            logger.debug(
                f"Failed to get MySQL row count estimate: {type(e).__name__}: {str(e)}"
            )
            return None



import logging
import re
from collections import Counter
from dataclasses import dataclass
from io import BytesIO
from typing import Any, Dict, Iterator, List, Optional, Union

import openpyxl
import pandas as pd
from openpyxl.workbook import Workbook

from datahub.ingestion.source.excel.report import ExcelSourceReport

logger = logging.getLogger(__name__)


@dataclass
class ExcelTable:
    df: pd.DataFrame
    header_row: int
    footer_row: int
    row_count: int
    column_count: int
    metadata: Dict[str, Any]
    sheet_name: str


class ExcelFile:
    wb: Workbook
    filename: str
    data: BytesIO
    sheet_list: List[str]
    active_sheet: str
    properties: Dict[str, Any]
    report: ExcelSourceReport

    def __init__(
        self,
        filename: str,
        data: BytesIO,
        report: ExcelSourceReport,
    ) -> None:
        self.filename = filename
        self.data = data
        self.report = report
        self.sheet_list = []
        self.active_sheet = ""
        self.properties = {}

    def load_workbook(self) -> bool:
        try:
            self.wb = openpyxl.load_workbook(self.data, data_only=True)
            self.properties = self.read_excel_properties(self.wb)
            self.sheet_list = self.wb.sheetnames
            self.active_sheet = self.wb.active.title
            return True
        except Exception as e:
            self.report.report_file_dropped(self.filename)
            self.report.warning(
                message="Error reading Excel file",
                context=f"Filename={self.filename}",
                exc=e,
            )
            return False

    @property
    def sheet_names(self) -> List[str]:
        return self.sheet_list

    @property
    def active_sheet_name(self) -> str:
        return self.active_sheet

    @property
    def workbook_properties(self) -> Dict[str, Any]:
        return self.properties

    def get_tables(self, active_only: Optional[bool] = False) -> Iterator[ExcelTable]:
        sheet_list = [self.active_sheet] if active_only else self.sheet_list
        for sheet in sheet_list:
            table = self.get_table(sheet)
            if table is not None:
                yield table
            else:
                self.report.report_worksheet_dropped(sheet)
                self.report.warning(
                    message="Worksheet does not contain a table",
                    context=f"Worksheet=[{self.filename}]{sheet}",
                )

    def get_table(self, sheet_name: str) -> Union[ExcelTable, None]:
        sheet = self.wb[sheet_name]

        # Extract all rows from the sheet
        rows = [[cell.value for cell in row] for row in sheet.rows]

        # Find a potential header row
        header_row_idx = self.find_header_row(rows)
        if header_row_idx is None:
            return None

        # Find where the footer starts
        footer_start_idx = self.find_footer_start(rows, header_row_idx)

        # Extract metadata before the header
        header_metadata = self.extract_metadata(rows[:header_row_idx])

        # Extract footer metadata
        footer_metadata = {}
        if footer_start_idx < len(rows):
            footer_metadata = self.extract_metadata(rows[footer_start_idx:])

        # Combine metadata
        metadata = {}
        metadata.update(self.properties)

        # Add header metadata
        for key, value in header_metadata.items():
            if key not in metadata:
                metadata[key] = value
            else:
                metadata[f"{key}_1"] = value

        # Add footer metadata
        for key, value in footer_metadata.items():
            if key not in metadata:
                metadata[key] = value
            else:
                metadata[f"{key}_1"] = value

        # Get the header row
        header_row = rows[header_row_idx]

        # Find the last non-empty column in the header row
        last_non_empty_idx = -1
        for i in range(len(header_row) - 1, -1, -1):
            if header_row[i] is not None and str(header_row[i]).strip() != "":
                last_non_empty_idx = i
                break

        # Truncate the header row to remove empty trailing columns
        if last_non_empty_idx >= 0:
            header_row = header_row[: last_non_empty_idx + 1]

        # Create the column names for the DataFrame
        column_names: List[str] = []
        seen_columns: Dict[str, int] = {}
        for i, col in enumerate(header_row):
            if col is None or str(col).strip() == "":
                col_name = f"Unnamed_{i}"
            else:
                col_name = str(col).strip()

            if col_name in seen_columns:
                seen_columns[col_name] += 1
                col_name = f"{col_name}_{seen_columns[col_name]}"
            else:
                seen_columns[col_name] = 0

            column_names.append(col_name)

        # Create the DataFrame with the table data
        data_rows = rows[header_row_idx + 1 : footer_start_idx]

        # Truncate data rows to match the header length
        truncated_data_rows = [
            row[: len(column_names)] if len(row) > len(column_names) else row
            for row in data_rows
        ]

        # Create the final DataFrame
        df = pd.DataFrame(truncated_data_rows, columns=column_names)

        row_count = df.shape[0]
        column_count = df.shape[1]

        return ExcelTable(
            df,
            header_row_idx + 1,
            footer_start_idx,
            row_count,
            column_count,
            metadata,
            sheet.title.strip(),
        )

    def find_header_row(self, rows: List[List[Any]]) -> Union[int, None]:
        max_score = -1
        header_idx = 0

        # Skip empty rows at the beginning
        start_idx = self._find_first_non_empty_row(rows)

        # Evaluate each potential header row with a lookahead
        min_rows_required = 2

        # Skip evaluation if there aren't enough rows
        if len(rows) < start_idx + min_rows_required + 1:
            return header_idx

        for i in range(start_idx, len(rows) - min_rows_required):
            current_row = rows[i]
            # Take as many next rows as available, up to 3
            next_rows = rows[i + 1 : min(i + 4, len(rows))]

            # Skip empty rows
            if not self._is_non_empty_row(current_row):
                continue

            score = self._calculate_header_row_score(i, current_row, next_rows, rows)

            if score > max_score:
                max_score = score
                header_idx = i

        if max_score <= 0:
            return None
        else:
            return header_idx

    def _find_first_non_empty_row(self, rows: List[List[Any]]) -> int:
        for i, row in enumerate(rows):
            if self._is_non_empty_row(row):
                return i
        return 0

    @staticmethod
    def _is_non_empty_row(row: List[Any]) -> bool:
        return any(cell is not None and str(cell).strip() != "" for cell in row)

    def _calculate_header_row_score(
        self,
        row_idx: int,
        current_row: List[Any],
        next_rows: List[List[Any]],
        all_rows: List[List[Any]],
    ) -> int:
        score = 0

        score += ExcelFile._score_row_with_numeric_cells(current_row)
        if score < 0:
            return score
        score += self._score_non_empty_cells(row_idx, current_row, all_rows)
        score += self._score_header_like_text(current_row)
        score += self._score_text_followed_by_numeric(current_row, next_rows)
        score += self._score_column_type_consistency(current_row, next_rows)
        score += self._score_metadata_patterns(row_idx, current_row, all_rows)

        return score

    @staticmethod
    def _score_non_empty_cells(
        row_idx: int, current_row: List[Any], all_rows: List[List[Any]]
    ) -> int:
        if row_idx <= 0:
            return 0

        non_empty_current = sum(
            1 for cell in current_row if cell is not None and str(cell).strip() != ""
        )

        non_empty_prev = sum(
            1
            for cell in all_rows[row_idx - 1]
            if cell is not None and str(cell).strip() != ""
        )

        return 2 if non_empty_current > non_empty_prev else 0

    @staticmethod
    def _score_header_like_text(row: List[Any]) -> int:
        return sum(
            1
            for cell in row
            if cell is not None
            and isinstance(cell, str)
            and re.match(r"^[A-Z][a-zA-Z\s]*$", str(cell).strip())
        )

    @staticmethod
    def _score_row_with_numeric_cells(row: List[Any]) -> int:
        return sum(
            -1 for cell in row if cell is not None and isinstance(cell, (int, float))
        )

    @staticmethod
    def _score_text_followed_by_numeric(
        current_row: List[Any], next_rows: List[List[Any]]
    ) -> int:
        if not next_rows:
            return 0

        header_text_count = sum(
            1 for cell in current_row if cell is not None and isinstance(cell, str)
        )

        next_rows_numeric_count = [
            sum(
                1
                for cell in row
                if cell is not None
                and (
                    isinstance(cell, (int, float))
                    or (
                        isinstance(cell, str)
                        and re.match(r"^-?\d+(\.\d+)?$", str(cell).strip())
                    )
                )
            )
            for row in next_rows
        ]

        if header_text_count > 0 and any(
            count > 0 for count in next_rows_numeric_count
        ):
            return 6 + sum(1 for count in next_rows_numeric_count if count > 0)
        return 0

    @staticmethod
    def _score_column_type_consistency(
        current_row: List[Any], next_rows: List[List[Any]]
    ) -> int:
        if len(next_rows) < 2:
            return 0

        col_types = []
        for col_idx in range(len(current_row)):
            if col_idx < len(current_row) and current_row[col_idx] is not None:
                col_type_counter: Counter = Counter()
                for row in next_rows:
                    if col_idx < len(row) and row[col_idx] is not None:
                        cell_type = type(row[col_idx]).__name__
                        col_type_counter[cell_type] += 1

                if col_type_counter and col_type_counter.most_common(1)[0][1] >= 1:
                    col_types.append(col_type_counter.most_common(1)[0][0])

        return 3 if len(col_types) >= 2 and len(set(col_types)) >= 1 else 0

    @staticmethod
    def _score_metadata_patterns(
        row_idx: int, current_row: List[Any], all_rows: List[List[Any]]
    ) -> int:
        score = 0

        if row_idx == 0 and len(current_row) <= 2:
            metadata_like = sum(
                1
                for cell in current_row
                if cell is not None and isinstance(cell, str) and len(str(cell)) <= 20
            )
            if metadata_like <= 2:
                score -= 1

        if row_idx < len(all_rows) - 1 and len(current_row) >= 2:
            if all(
                isinstance(cell, str) for cell in current_row[:2] if cell is not None
            ):
                score -= 2

        return score

    @staticmethod
    def find_footer_start(rows: List[List[Any]], header_row_idx: int) -> int:
        if header_row_idx + 1 >= len(rows):
            return len(rows)

        # Start with the assumption that all rows after the header are data (no footer)
        footer_start_idx = len(rows)

        # Get the number of columns in the header row to determine table width
        header_row = rows[header_row_idx]
        table_width = sum(
            1 for cell in header_row if cell is not None and str(cell).strip() != ""
        )

        # Get a sample of data rows to establish patterns
        data_sample_idx = min(header_row_idx + 5, len(rows) - 1)
        data_rows = rows[header_row_idx + 1 : data_sample_idx + 1]

        # Check for rows with significantly fewer populated cells than the data rows
        avg_populated_cells = sum(
            sum(1 for cell in row if cell is not None and str(cell).strip() != "")
            for row in data_rows
        ) / len(data_rows)

        # Look for pattern breaks, empty rows, or format changes
        for i in range(header_row_idx + 1, len(rows)):
            current_row = rows[i]

            # Skip completely empty rows unless followed by non-data-like rows
            if not any(
                cell is not None and str(cell).strip() != "" for cell in current_row
            ):
                # Look ahead to see if this empty row marks the start of footer
                if i + 1 < len(rows):
                    next_row = rows[i + 1]
                    next_row_populated = sum(
                        1
                        for cell in next_row
                        if cell is not None and str(cell).strip() != ""
                    )

                    # If the next row has significantly fewer populated cells or is text-heavy,
                    # consider this the start of footer
                    if (
                        next_row_populated < avg_populated_cells * 0.5
                        or sum(
                            1
                            for cell in next_row
                            if cell is not None
                            and isinstance(cell, str)
                            and len(str(cell)) > 20
                        )
                        > 0
                    ):
                        footer_start_idx = i
                        break
                continue

            # Count populated cells
            populated_cells = sum(
                1
                for cell in current_row
                if cell is not None and str(cell).strip() != ""
            )

            # Check for footer indicators
            footer_indicators = [
                "total",
                "sum",
                "average",
                "mean",
                "source",
                "note",
                "footnote",
            ]
            has_footer_text = any(
                cell is not None
                and isinstance(cell, str)
                and any(
                    indicator in str(cell).lower() for indicator in footer_indicators
                )
                for cell in current_row
            )

            # Check for the summary row
            looks_like_summary = has_footer_text and populated_cells <= table_width

            # Check for notes or sources (often longer text spanning multiple columns)
            long_text_cells = sum(
                1
                for cell in current_row
                if cell is not None and isinstance(cell, str) and len(str(cell)) > 50
            )

            # If this looks like the start of the footer, mark it
            if (
                (populated_cells < avg_populated_cells * 0.7 and i > header_row_idx + 3)
                or looks_like_summary
                or long_text_cells > 0
            ):
                footer_start_idx = i
                break

            # Check for inconsistent data types compared to data rows
            if i > header_row_idx + 3:
                data_type_mismatch = 0
                for j, cell in enumerate(current_row):
                    if j < len(header_row) and header_row[j] is not None:
                        # Get the most common data type for this column in previous rows
                        col_types = [
                            type(rows[idx][j]).__name__
                            for idx in range(header_row_idx + 1, i)
                            if idx < len(rows)
                            and j < len(rows[idx])
                            and rows[idx][j] is not None
                        ]
                        if col_types and cell is not None:
                            most_common_type = Counter(col_types).most_common(1)[0][0]
                            if type(cell).__name__ != most_common_type:
                                data_type_mismatch += 1

                # If many columns have type mismatches, this might be a footer row
                if data_type_mismatch > table_width * 0.5:
                    footer_start_idx = i
                    break

        return footer_start_idx

    @staticmethod
    def extract_metadata(rows: List[List[Any]]) -> Dict[str, Any]:
        metadata = {}

        for row in rows:
            if len(row) >= 2 and all(item is None for item in row[2:]):
                key, value = row[:2]
                if key is not None and value is not None:
                    metadata[str(key).strip().rstrip(":=").rstrip()] = str(
                        value
                    ).strip()

        return metadata

    @staticmethod
    def read_excel_properties(wb: Workbook) -> Dict[str, Any]:
        # Core properties from DocumentProperties
        core_props = wb.properties
        properties = {
            "title": core_props.title,
            "author": core_props.creator,
            "subject": core_props.subject,
            "description": core_props.description,
            "keywords": core_props.keywords,
            "category": core_props.category,
            "last_modified_by": core_props.lastModifiedBy,
            "created": core_props.created,
            "modified": core_props.modified,
            "status": core_props.contentStatus,
            "revision": core_props.revision,
            "version": core_props.version,
            "language": core_props.language,
            "identifier": core_props.identifier,
        }

        # Remove None values
        properties = {k: v for k, v in properties.items() if v is not None}

        # Assign custom properties if they exist
        if hasattr(wb, "custom_doc_props"):
            for prop in wb.custom_doc_props.props:
                if prop.value:
                    if prop.name in properties:
                        prop_name = f"custom.{prop.name}"
                    else:
                        prop_name = prop.name
                    properties[prop_name] = prop.value

        return properties



"""
Utilities for navigating a powerquery-parser NodeIdMap.

A NodeIdMap is dict[int, dict] — the deserialized output of the TypeScript
bridge. Each node dict has at minimum:
    "kind": str   — NodeKind enum value (e.g. "LetExpression")
    "id": int     — unique node identifier

Child nodes are embedded directly as nested dicts, not as integer ID references.
The nodeIdMap provides a flat index to find any node by ID.
"""

from typing import Dict, Optional

NodeIdMap = dict[int, dict]


def find_nodes_by_kind(node_map: NodeIdMap, kind: str) -> list[dict]:
    """Return all nodes of the given NodeKind at any depth in the map."""
    return [node for node in node_map.values() if node.get("kind") == kind]


def get_literal_value(node: dict) -> Optional[str]:
    """
    Return the string content of a Text LiteralExpression with quotes stripped.
    Returns None for non-Text literals or non-LiteralExpression nodes.
    """
    if node.get("kind") != "LiteralExpression":
        return None
    if node.get("literalKind") != "Text":
        return None
    literal = node.get("literal", "")
    if not (literal.startswith('"') and literal.endswith('"')):
        return None
    return literal[1:-1]


def get_invoke_callee_name(node_map: NodeIdMap, invoke_node: dict) -> Optional[str]:
    """
    Resolve the callee name of an InvokeExpression.

    The powerquery-parser AST encodes ``Snowflake.Databases(...)`` as::

        RecursivePrimaryExpression
          head: IdentifierExpression  → identifier.literal = "Snowflake.Databases"
          recursiveExpressions: ArrayWrapper
            elements: [InvokeExpression ← this is the node passed in]

    The function name lives on the *parent*, not the invoke node itself, so we
    scan the flat map for a RecursivePrimaryExpression whose element list
    contains this invoke node's ID.

    Returns None if the node is not an InvokeExpression or the callee cannot be resolved.
    """
    if invoke_node.get("kind") != "InvokeExpression":
        return None

    invoke_id = invoke_node.get("id")
    if invoke_id is None:
        return None

    # Find the parent RecursivePrimaryExpression that contains this InvokeExpression
    for node in node_map.values():
        if node.get("kind") != "RecursivePrimaryExpression":
            continue
        rec_exprs = node.get("recursiveExpressions", {})
        if not isinstance(rec_exprs, dict) or rec_exprs.get("kind") != "ArrayWrapper":
            continue
        for elem in rec_exprs.get("elements", []):
            if isinstance(elem, dict) and elem.get("id") == invoke_id:
                head = node.get("head", {})
                if head.get("kind") == "IdentifierExpression":
                    identifier = head.get("identifier", {})
                    return identifier.get("literal")

    return None


def resolve_identifier(
    node_map: NodeIdMap,
    let_node: dict,
    name: str,
) -> Optional[dict]:
    """
    Look up a variable name in the given LetExpression's variable list.
    Returns the value node assigned to `name`, or None if not found.

    Structure: LetExpression.variableList (ArrayWrapper) → elements (Csv[]) → node (IdentifierPairedExpression)
    """
    if let_node.get("kind") != "LetExpression":
        return None

    var_list = let_node.get("variableList", {})
    if not isinstance(var_list, dict) or var_list.get("kind") != "ArrayWrapper":
        return None

    for elem in var_list.get("elements", []):
        # elements are Csv nodes wrapping IdentifierPairedExpression
        if isinstance(elem, dict) and elem.get("kind") == "Csv":
            inner = elem.get("node", {})
        else:
            inner = elem

        if not isinstance(inner, dict):
            continue
        if inner.get("kind") not in (
            "IdentifierPairedExpression",
            "GeneralizedIdentifierPairedExpression",
        ):
            continue

        key_node = inner.get("key", {})
        if not isinstance(key_node, dict):
            continue
        key_literal = key_node.get("literal", "")
        # Normalize: strip #"..." quoting and compare case-insensitively
        # (M-Query variable names are case-insensitive)
        key_bare = key_literal
        if key_bare.startswith('#"') and key_bare.endswith('"'):
            key_bare = key_bare[2:-1]
        name_bare = name
        if name_bare.startswith('#"') and name_bare.endswith('"'):
            name_bare = name_bare[2:-1]
        if key_bare.lower() == name_bare.lower():
            return inner.get("value")

    return None


def get_record_field_values(
    node_map: NodeIdMap,
    record_node: dict,
    parameters: Optional[Dict[str, str]] = None,
) -> dict[str, str]:
    """
    Extract key-value pairs from a RecordExpression where values are Text literals.
    Keys: GeneralizedIdentifier literals.
    Values: Text LiteralExpression values (quotes stripped).
    Non-string values are omitted unless resolvable via parameters.

    Structure: RecordExpression.content (ArrayWrapper) → elements (Csv[]) → node (GeneralizedIdentifierPairedExpression)
    """
    parameters = parameters or {}
    result: dict[str, str] = {}
    if record_node.get("kind") != "RecordExpression":
        return result

    content = record_node.get("content", {})
    if not isinstance(content, dict) or content.get("kind") != "ArrayWrapper":
        return result

    for elem in content.get("elements", []):
        if isinstance(elem, dict) and elem.get("kind") == "Csv":
            inner = elem.get("node", {})
        else:
            inner = elem

        if not isinstance(inner, dict):
            continue
        if inner.get("kind") not in (
            "GeneralizedIdentifierPairedExpression",
            "IdentifierPairedExpression",
        ):
            continue

        key_node = inner.get("key", {})
        value_node = inner.get("value", {})
        if not isinstance(key_node, dict) or not isinstance(value_node, dict):
            continue

        key = key_node.get("literal", "")
        value = get_literal_value(value_node)
        if (
            value is None
            and parameters
            and value_node.get("kind") == "IdentifierExpression"
        ):
            # Resolve identifier references using parameters
            ident = value_node.get("identifier", {})
            ref_name = ident.get("literal", "")
            if ref_name.startswith('#"') and ref_name.endswith('"'):
                ref_name = ref_name[2:-1]
            if ref_name in parameters:
                value = parameters[ref_name]
        if value is not None:
            result[key] = value

    return result



"""
UA Parser enrichment lineage extractor.

Extracts field-level lineage for the UA Parser enrichment, which parses the user agent
string to extract browser, OS, and device information.

Enrichment Documentation:
https://docs.snowplow.io/docs/pipeline/enrichments/available-enrichments/ua-parser-enrichment/

Schema:
iglu:com.snowplowanalytics.snowplow/ua_parser_context/jsonschema/1-0-0

Field Mapping Reference:
/connectors-accelerator/SNOWPLOW_ENRICHMENT_FIELD_MAPPING.md#3-ua-parser-enrichment

Note:
This enrichment outputs to a CONTEXT (not atomic fields). Context lineage tracking
is planned for a future enhancement. For now, this extractor returns empty lineages.

Output Context Fields (not yet tracked):
- useragent_family, useragent_major, useragent_minor, useragent_patch, useragent_version
- os_family, os_major, os_minor, os_patch, os_patch_minor, os_version
- device_family
"""

import logging
from typing import List, Optional

from datahub.ingestion.source.snowplow.enrichment_lineage.base import (
    EnrichmentFieldInfo,
    EnrichmentLineageExtractor,
    FieldLineage,
)
from datahub.ingestion.source.snowplow.models.snowplow_models import Enrichment

logger = logging.getLogger(__name__)


class UaParserLineageExtractor(EnrichmentLineageExtractor):
    """
    Extractor for UA Parser enrichment lineage.

    Input Fields:
    - useragent (atomic field from HTTP User-Agent header)

    Output Fields:
    - Context: com.snowplowanalytics.snowplow/ua_parser_context/jsonschema/1-0-0
      Contains 12 fields: useragent_family, useragent_major, useragent_minor, etc.

    Current Limitation:
    Context lineage is not yet implemented. This enrichment adds a derived context
    rather than atomic event fields. Future enhancement will track lineage to
    warehouse context tables (e.g., events_contexts_com_snowplowanalytics_snowplow_ua_parser_context_1).
    """

    def supports_enrichment(self, enrichment_schema: str) -> bool:
        """
        Check if this extractor handles UA Parser enrichments.

        Args:
            enrichment_schema: Iglu schema URI

        Returns:
            True if schema is for UA Parser enrichment
        """
        # The enrichment config schema, not the output context schema
        return "ua_parser_config" in enrichment_schema

    def extract_lineage(
        self,
        enrichment: Enrichment,
        event_schema_urn: str,
        warehouse_table_urn: Optional[str],
    ) -> List[FieldLineage]:
        """
        Extract field-level lineage for UA Parser enrichment.

        Currently returns empty list because context lineage is not yet implemented.

        Args:
            enrichment: UA Parser enrichment configuration
            event_schema_urn: URN of input event schema
            warehouse_table_urn: URN of output warehouse table

        Returns:
            Empty list (context lineage not yet supported)
        """
        # TODO: Implement context lineage tracking
        # Would need to:
        # 1. Construct context table URN (e.g., warehouse.events_contexts_com_snowplowanalytics_snowplow_ua_parser_context_1)
        # 2. Map useragent → context fields (useragent_family, os_family, device_family, etc.)
        # 3. Return FieldLineage objects for each field

        logger.debug(
            "UA Parser: Context lineage not yet implemented, returning empty lineages"
        )
        return []

    def get_field_info(self, enrichment: Enrichment) -> EnrichmentFieldInfo:
        """
        Get field information for UA Parser enrichment.

        Note: This enrichment outputs to a context table, not atomic fields.
        The field info is still useful for description building.
        """
        return EnrichmentFieldInfo(
            input_fields=["useragent"],
            output_fields=[
                "useragent_family",
                "useragent_major",
                "useragent_minor",
                "os_family",
                "os_major",
                "os_minor",
                "device_family",
            ],
            transformation_description="Parses user agent string (outputs to context table)",
        )



import os
from typing import Dict, List, Union

from datahub.secret.secret_store import SecretStore


# Simple SecretStore implementation that fetches Secret values from the local environment.
class EnvironmentSecretStore(SecretStore):
    def __init__(self, config):
        pass

    def close(self) -> None:
        return

    def get_secret_values(self, secret_names: List[str]) -> Dict[str, Union[str, None]]:
        values = {}
        for secret_name in secret_names:
            values[secret_name] = os.getenv(secret_name)
        return values

    def get_secret_value(self, secret_name: str) -> Union[str, None]:
        return os.getenv(secret_name)

    def get_id(self) -> str:
        return "env"

    @classmethod
    def create(cls, config: Dict) -> "EnvironmentSecretStore":
        return cls(config)



"""Report classes for Fabric clients."""

from dataclasses import dataclass, field

from datahub.utilities.lossy_collections import LossyList


@dataclass
class FabricClientReport:
    """Report for Fabric REST API client operations.

    Tracks metrics for API calls, errors, and other client-level operations.
    This can be aggregated into workload-specific source reports.
    """

    request_count: int = 0
    error_count: int = 0
    api_parse_failures: LossyList[str] = field(default_factory=LossyList)

    def report_request(self) -> None:
        """Track a successful API request."""
        self.request_count += 1

    def report_error(self) -> None:
        """Track an API error."""
        self.error_count += 1

    def report_parse_failure(self, context: str) -> None:
        """Track an API response that couldn't be parsed due to missing/unexpected fields."""
        self.api_parse_failures.append(context)



"""Configuration for Google Dataplex source."""

import logging
from typing import Dict, List, Optional

from pydantic import Field, field_validator, model_validator

from datahub.configuration.common import AllowDenyPattern, ConfigModel
from datahub.configuration.source_common import (
    EnvConfigMixin,
    PlatformInstanceConfigMixin,
)
from datahub.ingestion.source.common.gcp_credentials_config import GCPCredential
from datahub.ingestion.source.common.gcp_project_filter import (
    GcpProjectFilterConfig,
    GCPValidationError,
    validate_project_label_list,
)
from datahub.ingestion.source.state.stale_entity_removal_handler import (
    StatefulStaleMetadataRemovalConfig,
)
from datahub.ingestion.source.state.stateful_ingestion_base import (
    StatefulIngestionConfigBase,
    StatefulLineageConfigMixin,
)

logger = logging.getLogger(__name__)

DEFAULT_LINEAGE_LOCATIONS = [
    "us",
    "eu",
    "asia",
    "us-central1",
    "us-east1",
    "us-east4",
    "us-east5",
    "us-south1",
    "us-west1",
    "us-west2",
    "us-west3",
    "us-west4",
    "northamerica-northeast1",
    "northamerica-northeast2",
    "southamerica-east1",
    "southamerica-west1",
    "europe-central2",
    "europe-north1",
    "europe-southwest1",
    "europe-west1",
    "europe-west2",
    "europe-west3",
    "europe-west4",
    "europe-west6",
    "europe-west8",
    "europe-west9",
    "europe-west10",
    "europe-west12",
    "me-central1",
    "me-central2",
    "me-west1",
    "asia-east1",
    "asia-east2",
    "asia-northeast1",
    "asia-northeast2",
    "asia-northeast3",
    "asia-south1",
    "asia-south2",
    "asia-southeast1",
    "asia-southeast2",
    "australia-southeast1",
    "australia-southeast2",
    "africa-south1",
]


class EntriesFilterConfig(ConfigModel):
    """Filter configuration specific to Dataplex Entries API (Universal Catalog)."""

    pattern: AllowDenyPattern = Field(
        default=AllowDenyPattern.allow_all(),
        description="Regex patterns for Dataplex entry names to filter in ingestion.",
    )
    fqn_pattern: AllowDenyPattern = Field(
        default=AllowDenyPattern.allow_all(),
        description="Regex patterns for Dataplex fully-qualified names to filter in ingestion.",
    )


class EntryGroupFilterConfig(ConfigModel):
    """Filter configuration for Dataplex entry groups."""

    pattern: AllowDenyPattern = Field(
        default=AllowDenyPattern.allow_all(),
        description="Regex patterns for entry group resource names to include/exclude.",
    )


class DataplexFilterConfig(ConfigModel):
    """Filter configuration for Dataplex ingestion."""

    entry_groups: EntryGroupFilterConfig = Field(
        default_factory=EntryGroupFilterConfig,
        description="Filters for Dataplex entry group names.",
    )
    entries: EntriesFilterConfig = Field(
        default_factory=EntriesFilterConfig,
        description="Filters specific to Dataplex Entries API (Universal Catalog).",
    )


class DataplexConfig(
    GcpProjectFilterConfig,
    EnvConfigMixin,
    PlatformInstanceConfigMixin,
    StatefulIngestionConfigBase,
    StatefulLineageConfigMixin,
):
    """Configuration for Google Dataplex source.

    Project selection (`project_ids`, `project_labels`, `project_id_pattern`) is
    inherited from `GcpProjectFilterConfig` and consumed by the shared
    `resolve_gcp_projects` helper. Auto-discovery (when `project_ids` is empty)
    requires `roles/resourcemanager.folderViewer` on the parent folder/organization.
    """

    credential: Optional[GCPCredential] = Field(
        default=None,
        description="GCP credential information. If not specified, uses Application Default Credentials.",
    )

    entries_locations: List[str] = Field(
        default_factory=lambda: ["us", "eu", "asia", "global"],
        description="List of GCP regions to scan for Universal Catalog entries extraction. "
        "This list may include multi-regions (for example 'us', 'eu', 'asia') and "
        "single regions (for example 'us-central1'). "
        "Entries scanning runs across all configured entries_locations. "
        "Default: ['us', 'eu', 'asia', 'global'].",
    )

    filter_config: DataplexFilterConfig = Field(
        default_factory=DataplexFilterConfig,
        description="Filters to control which Dataplex resources are ingested.",
    )

    include_schema: bool = Field(
        default=True,
        description="Whether to extract and ingest schema metadata (columns, types, descriptions). "
        "Set to False to skip schema extraction for faster ingestion when only basic dataset metadata is needed. "
        "Disabling schema extraction can improve performance for large deployments. Default: True.",
    )

    include_lineage: bool = Field(
        default=True,
        description="Whether to extract lineage information using Dataplex Lineage API. "
        "Extracts table-level lineage relationships between entries. "
        "Lineage API calls automatically retry transient errors (timeouts, rate limits) with exponential backoff.",
    )

    lineage_locations: List[str] = Field(
        default_factory=lambda: list(DEFAULT_LINEAGE_LOCATIONS),
        description="List of GCP regions to scan for Dataplex lineage data. "
        "By default, includes all supported multi-regions and regions. "
        "Narrowing this list from the default is critical for better performance "
        "because lineage API calls scale with configured project/location pairs. "
        "This list may include multi-regions and single regions. "
        "In practice, lineage often resides in job regions while entries may be in "
        "multi-regions, so entries_locations and lineage_locations are configured separately. "
        "Example: ['eu', 'us-central1', 'europe-west1'].",
    )

    lineage_max_retries: int = Field(
        default=3,
        ge=1,
        le=10,
        description="Maximum number of retry attempts for lineage API calls when encountering transient errors "
        "(timeouts, rate limits, service unavailable). Each attempt uses exponential backoff. "
        "Higher values increase resilience but may slow down ingestion. Default: 3.",
    )

    lineage_retry_backoff_multiplier: float = Field(
        default=1.0,
        ge=0.1,
        le=10.0,
        description="Multiplier for exponential backoff between lineage API retry attempts (in seconds). "
        "Wait time formula: multiplier * (2 ^ attempt_number), capped between 2-10 seconds. "
        "Higher values reduce API load but increase ingestion time. Default: 1.0.",
    )

    max_workers_entries: int = Field(
        default=10,
        ge=1,
        le=100,
        description="Number of parallel worker threads for fetching entry details "
        "(get_entry API calls). Entry detail fetching is the main bottleneck in the "
        "entries stage because each entry requires one blocking RPC. Increasing this "
        "value reduces wall-clock time proportionally up to the API quota limit. "
        "Increase for large deployments (>1k entries). Default: 10.",
    )

    max_workers_lineage: int = Field(
        default=10,
        ge=1,
        le=100,
        description="Number of parallel worker threads for lineage lookups "
        "(search_links API calls). Lineage lookup volume scales with entries × "
        "lineage_locations, so parallelism here has a large impact on total "
        "ingestion time. Increase for large entry × location matrices. Default: 10.",
    )

    stateful_ingestion: Optional[StatefulStaleMetadataRemovalConfig] = Field(
        default=None,
        description="Stateful ingestion configuration for stale metadata removal.",
    )

    include_glossaries: bool = Field(
        default=True,
        description=(
            "Whether to ingest Dataplex Business Glossary entities as DataHub GlossaryNodes "
            "and GlossaryTerms. Glossaries, categories, and terms are emitted with correct "
            "parent hierarchy. Default: True."
        ),
    )

    include_glossary_term_associations: bool = Field(
        default=False,
        description=(
            "Whether to ingest term-to-asset associations via the Dataplex lookupEntryLinks API. "
            "For each ingested term, all entries_locations are queried per project to find linked "
            "assets. Requires roles/resourcemanager.projectViewer on all configured projects to "
            "resolve GCP project numbers needed by the lookupEntryLinks API."
        ),
    )

    glossary_locations: List[str] = Field(
        default_factory=lambda: ["global"],
        description=(
            "GCP locations to scan for Dataplex Business Glossaries. "
            "Dataplex glossaries are typically created in 'global' but can exist in any location. "
            "Default: ['global']."
        ),
    )

    max_workers_glossary: int = Field(
        default=10,
        ge=1,
        le=100,
        description=(
            "Number of parallel worker threads for glossary ingestion (fetching terms and "
            "categories per glossary) and term-asset association traversal "
            "(lookupEntryLinks calls). Default: 10."
        ),
    )

    dataplex_url: str = Field(
        default="https://console.cloud.google.com/dataplex",
        description="Base URL for Dataplex console (for generating external links).",
    )

    @model_validator(mode="before")
    @classmethod
    def project_id_backward_compatibility(cls, values: Dict) -> Dict:
        """Handle backward compatibility for project_id -> project_ids migration."""
        # Pydantic passes the raw input dict to mode="before" validators.
        # We create a new dict to avoid mutating the input (important for dict reuse in tests).
        project_id = values.get("project_id")
        project_ids = values.get("project_ids")

        if not project_ids and project_id:
            # Create a new dict without project_id, adding project_ids
            result = {k: v for k, v in values.items() if k != "project_id"}
            result["project_ids"] = [project_id]
            return result
        elif project_ids and project_id:
            logging.warning(
                "Both project_id and project_ids are set. Using project_ids. "
                "The project_id config is deprecated, please use project_ids instead."
            )
            # Remove project_id from the dict
            return {k: v for k, v in values.items() if k != "project_id"}

        return values

    @field_validator("project_labels")
    @classmethod
    def _validate_project_labels_format(cls, v: List[str]) -> List[str]:
        try:
            validate_project_label_list(v)
        except GCPValidationError as e:
            raise ValueError(str(e)) from e
        return v

    @model_validator(mode="after")
    def validate_project_ids(self) -> "DataplexConfig":
        """Ensure at least one means of selecting projects is configured."""
        has_non_default_pattern = (
            self.project_id_pattern != AllowDenyPattern.allow_all()
        )
        if (
            not self.project_ids
            and not self.project_labels
            and not has_non_default_pattern
        ):
            raise ValueError(
                "At least one project selector must be specified. Set project_ids "
                "explicitly, or use project_id_pattern / project_labels to "
                "auto-discover projects."
            )
        return self

    @model_validator(mode="after")
    def validate_location_configuration(self) -> "DataplexConfig":
        """Validate location configuration and warn about common mistakes."""
        if not self.entries_locations:
            raise ValueError(
                "At least one entries location must be specified via entries_locations."
            )
        if not self.lineage_locations:
            raise ValueError(
                "At least one lineage location must be specified via lineage_locations."
            )
        if self.include_glossaries and not self.glossary_locations:
            raise ValueError(
                "At least one glossary location must be specified via glossary_locations "
                "when include_glossaries is enabled."
            )

        return self

    def get_credentials(self) -> Optional[Dict[str, str]]:
        """Get credentials dictionary for authentication."""
        if self.credential:
            # Use the first project_id for credential context
            project_id = self.project_ids[0] if self.project_ids else None
            return self.credential.to_dict(project_id)
        return None



"""
Deployment fetching service for Snowplow connector.

Handles fetching deployment history for data structures from BDP API,
supporting both parallel and sequential fetching strategies.
"""

import logging
from concurrent.futures import ThreadPoolExecutor, as_completed
from typing import TYPE_CHECKING, Dict, List, Optional, Tuple

from datahub.ingestion.source.snowplow.models.snowplow_models import (
    DataStructure,
    DataStructureDeployment,
)

if TYPE_CHECKING:
    from datahub.ingestion.source.snowplow.snowplow_client import SnowplowBDPClient
    from datahub.ingestion.source.snowplow.snowplow_report import SnowplowSourceReport

logger = logging.getLogger(__name__)


class DeploymentFetcher:
    """
    Service for fetching deployment history from Snowplow BDP.

    Supports parallel fetching for improved performance when processing
    large numbers of schemas.
    """

    def __init__(
        self,
        bdp_client: Optional["SnowplowBDPClient"],
        enable_parallel: bool = True,
        max_workers: int = 5,
        report: Optional["SnowplowSourceReport"] = None,
    ) -> None:
        """
        Initialize deployment fetcher.

        Args:
            bdp_client: BDP API client instance
            enable_parallel: Whether to use parallel fetching
            max_workers: Maximum concurrent API calls for parallel fetching
            report: Optional report for tracking fetch failures
        """
        self.bdp_client = bdp_client
        self.enable_parallel = enable_parallel
        self.max_workers = max_workers
        self.report = report

    def fetch_deployments(self, data_structures: List[DataStructure]) -> None:
        """
        Fetch deployment history for all data structures.

        Mutates the data_structures list in place, adding deployments to each.

        Args:
            data_structures: List of data structures to fetch deployments for
        """
        if not self.bdp_client:
            logger.warning("No BDP client configured, skipping deployment fetching")
            return

        # Filter to schemas that have a hash (needed for API call)
        schemas_needing_deployments = [ds for ds in data_structures if ds.hash]

        if not schemas_needing_deployments:
            logger.debug("No schemas need deployment fetching")
            return

        logger.info(
            f"Fetching deployment history for {len(schemas_needing_deployments)} schemas"
        )

        if self.enable_parallel and len(schemas_needing_deployments) > 1:
            self._fetch_parallel(schemas_needing_deployments)
        else:
            self._fetch_sequential(schemas_needing_deployments)

    def _fetch_parallel(self, schemas: List[DataStructure]) -> None:
        """
        Fetch deployments in parallel using ThreadPoolExecutor.

        Uses a two-phase approach to avoid race conditions:
        1. Fetch all deployments concurrently, collecting results in a map
        2. Apply results to data structures in single thread

        Args:
            schemas: List of data structures to fetch deployments for
        """
        logger.info(
            f"Fetching deployments in parallel (max_workers={self.max_workers}) "
            f"for {len(schemas)} schemas"
        )

        def fetch_one(
            ds: DataStructure,
        ) -> Tuple[
            str, str, Optional[List[DataStructureDeployment]], Optional[Exception]
        ]:
            """
            Fetch deployments for a single schema without mutating.

            Returns:
                Tuple of (vendor, name, deployments, error)
            """
            try:
                if self.bdp_client and ds.hash:
                    deployments = self.bdp_client.get_data_structure_deployments(
                        ds.hash
                    )
                    logger.debug(
                        f"Fetched {len(deployments) if deployments else 0} deployments "
                        f"for {ds.vendor}/{ds.name}"
                    )
                    return ds.vendor or "", ds.name or "", deployments, None
                return ds.vendor or "", ds.name or "", None, None
            except Exception as e:
                return ds.vendor or "", ds.name or "", None, e

        # Phase 1: Fetch all deployments concurrently
        deployment_map: Dict[str, Optional[List[DataStructureDeployment]]] = {}
        with ThreadPoolExecutor(max_workers=self.max_workers) as executor:
            futures = [executor.submit(fetch_one, ds) for ds in schemas]

            failed_schemas: List[str] = []
            for future in as_completed(futures):
                try:
                    vendor, name, deployments, error = future.result()
                except Exception as e:
                    logger.warning(f"Unexpected error from deployment fetch task: {e}")
                    if self.report:
                        self.report.report_deployment_fetch_failure("unknown", str(e))
                    continue
                schema_key = f"{vendor}/{name}"
                if error:
                    failed_schemas.append(schema_key)
                    logger.warning(
                        f"Failed to fetch deployments for {schema_key}: {error}"
                    )
                    if self.report:
                        self.report.report_deployment_fetch_failure(
                            schema_key, str(error)
                        )
                else:
                    deployment_map[schema_key] = deployments

            if failed_schemas:
                logger.warning(
                    f"Failed to fetch deployments for {len(failed_schemas)} schema(s). "
                    f"These schemas will be extracted without deployment history."
                )

        # Phase 2: Apply results to schemas in single thread
        for ds in schemas:
            schema_key = f"{ds.vendor}/{ds.name}"
            if schema_key in deployment_map:
                deployments = deployment_map[schema_key]
                if deployments is not None:
                    ds.deployments = deployments

        logger.info(
            f"Completed parallel deployment fetching for {len(schemas)} schemas "
            f"({len(deployment_map)} successful)"
        )

    def _fetch_sequential(self, schemas: List[DataStructure]) -> None:
        """
        Fetch deployments sequentially.

        Args:
            schemas: List of data structures to fetch deployments for
        """
        if not self.bdp_client:
            return

        for ds in schemas:
            try:
                if ds.hash:
                    deployments = self.bdp_client.get_data_structure_deployments(
                        ds.hash
                    )
                    ds.deployments = deployments
                    logger.debug(
                        f"Fetched {len(deployments)} deployments for {ds.vendor}/{ds.name}"
                    )
            except Exception as e:
                schema_key = f"{ds.vendor}/{ds.name}"
                logger.warning(f"Failed to fetch deployments for {schema_key}: {e}")
                if self.report:
                    self.report.report_deployment_fetch_failure(schema_key, str(e))



"""
Snowplow BDP Console API client.

Handles authentication, API requests, retry logic, and pagination for the
Snowplow BDP (Behavioral Data Platform) Console API.

API Documentation: https://console.snowplowanalytics.com/api/msc/v1/docs/
"""

import json
import logging
import time
from typing import TYPE_CHECKING, Dict, List, Optional

import requests
from pydantic import ValidationError
from requests.adapters import HTTPAdapter
from urllib3.util.retry import Retry

if TYPE_CHECKING:
    from datahub.ingestion.source.snowplow.snowplow_report import SnowplowSourceReport

from datahub.ingestion.source.snowplow.models.snowplow_models import (
    DataModel,
    DataStructure,
    DataStructureDeployment,
    Destination,
    Enrichment,
    EventSpecification,
    EventSpecificationsResponse,
    Organization,
    Pipeline,
    PipelinesResponse,
    TokenResponse,
    TrackingPlan,
    TrackingPlansResponse,
    User,
)
from datahub.ingestion.source.snowplow.snowplow_config import (
    SnowplowBDPConnectionConfig,
)

logger = logging.getLogger(__name__)


class ResourceNotFoundError(Exception):
    """Raised when API resource returns 404 Not Found."""

    pass


class SnowplowBDPClient:
    """
    API client for Snowplow BDP Console.

    Handles:
    - v3 Authentication (API Key → JWT token)
    - Data structures (schemas) extraction
    - Event specifications extraction
    - Tracking scenarios extraction
    - Automatic token refresh on expiry
    - Retry logic with exponential backoff
    - Error handling
    """

    def __init__(
        self,
        config: SnowplowBDPConnectionConfig,
        report: Optional["SnowplowSourceReport"] = None,
    ):
        """
        Initialize BDP Console API client.

        Args:
            config: BDP connection configuration
            report: Optional report for tracking API metrics
        """
        self.config = config
        self.base_url = config.console_api_url
        self.organization_id = config.organization_id
        self.report = report  # For API metrics tracking

        # JWT token (obtained via authentication)
        self._jwt_token: Optional[str] = None

        # Setup session with retry logic
        self.session = requests.Session()
        retry_strategy = Retry(
            total=config.max_retries,
            backoff_factor=1,  # Exponential backoff: 1, 2, 4 seconds
            status_forcelist=[429, 500, 502, 503, 504],
            allowed_methods=["GET", "POST", "PUT"],  # Retry safe methods
        )
        adapter = HTTPAdapter(max_retries=retry_strategy)
        self.session.mount("http://", adapter)
        self.session.mount("https://", adapter)

        # Authenticate immediately
        self._authenticate()

    def close(self) -> None:
        """Close the HTTP session and release resources."""
        if hasattr(self, "session") and self.session:
            self.session.close()

    def __enter__(self) -> "SnowplowBDPClient":
        """Context manager entry."""
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        """Context manager exit - cleanup resources."""
        self.close()

    def _authenticate(self) -> None:
        """
        Authenticate with BDP Console API using v3 flow.

        Exchanges API credentials for JWT token. Credentials (api_key_id, api_key) are
        sent as headers, and the server returns a short-lived JWT for subsequent requests.

        API Reference: https://docs.snowplow.io/docs/using-the-snowplow-console/managing-console-api-authentication/

        The token endpoint uses GET (not POST). This is Snowplow's API convention,
        confirmed by snowplow-cli source code. Credentials are passed via
        X-API-Key-ID and X-API-Key headers rather than a request body.
        """
        # GET /organizations/{organizationId}/credentials/v3/token
        url = (
            f"{self.base_url}/organizations/{self.organization_id}/credentials/v3/token"
        )

        headers = {
            "X-API-Key-ID": self.config.api_key_id,
            "X-API-Key": self.config.api_key.get_secret_value(),
        }

        try:
            logger.info(f"Authenticating with Snowplow BDP Console API: {url}")
            response = self.session.get(
                url,
                headers=headers,
                timeout=self.config.timeout_seconds,
            )
            response.raise_for_status()

            # Parse response using Pydantic model
            response_data = response.json()
            try:
                token_response = TokenResponse.model_validate(response_data)
                self._jwt_token = token_response.access_token

                # Set JWT in session headers for future requests
                self.session.headers.update(
                    {"Authorization": f"Bearer {self._jwt_token}"}
                )

                logger.info("Successfully authenticated with Snowplow BDP Console API")
            except ValidationError as parse_error:
                logger.error(f"Failed to parse authentication response: {parse_error}")
                logger.debug(
                    f"Raw response data: {json.dumps(response_data, indent=2, default=str)}"
                )
                raise ValueError(
                    f"Invalid authentication response format: {parse_error}"
                ) from parse_error

        except requests.exceptions.HTTPError as e:
            if e.response.status_code == 401:
                raise ValueError(
                    f"Authentication failed: Invalid API credentials. "
                    f"Check api_key_id and api_key. Status: {e.response.status_code}"
                ) from e
            elif e.response.status_code == 403:
                raise ValueError(
                    f"Authentication failed: Forbidden. Check organization_id and API key permissions. "
                    f"Status: {e.response.status_code}"
                ) from e
            else:
                raise ValueError(
                    f"Authentication failed with status {e.response.status_code}: {e}"
                ) from e
        except requests.exceptions.RequestException as e:
            raise ValueError(f"Authentication failed due to network error: {e}") from e
        except json.JSONDecodeError as e:
            raise ValueError(
                f"Authentication failed: invalid JSON response: {e}"
            ) from e

    def _request(
        self,
        method: str,
        endpoint: str,
        params: Optional[dict] = None,
        json_data: Optional[dict] = None,
        retry_auth: bool = True,
    ) -> dict:
        """
        Make authenticated API request with error handling and timing metrics.

        Args:
            method: HTTP method (GET, POST, etc.)
            endpoint: API endpoint path (without base URL)
            params: Query parameters
            json_data: JSON request body
            retry_auth: Whether to retry with re-authentication on 401

        Returns:
            Response JSON as dictionary

        Raises:
            requests.exceptions.HTTPError: For HTTP errors
            ValueError: For authentication errors
        """
        url = f"{self.base_url}/{endpoint.lstrip('/')}"
        endpoint_name = self._extract_endpoint_name(endpoint)
        start_time = time.perf_counter()
        is_error = False

        try:
            response = self.session.request(
                method=method,
                url=url,
                params=params,
                json=json_data,
                timeout=self.config.timeout_seconds,
            )
            response.raise_for_status()
            return response.json()

        except requests.exceptions.HTTPError as e:
            if e.response.status_code == 401 and retry_auth:
                # JWT token expired, re-authenticate and retry
                logger.warning("JWT token expired, re-authenticating...")
                self._authenticate()
                return self._request(
                    method, endpoint, params, json_data, retry_auth=False
                )
            elif e.response.status_code == 403:
                is_error = True
                logger.error(f"Permission denied: {url}")
                raise PermissionError(
                    f"Permission denied for {url}. Check API key permissions."
                ) from e
            elif e.response.status_code == 404:
                # 404 is not always an error (e.g., checking if resource exists)
                logger.debug(f"Resource not found: {url}")
                raise ResourceNotFoundError(f"Resource not found: {url}") from e
            else:
                is_error = True
                logger.error(f"HTTP error {e.response.status_code}: {url}")
                raise
        except requests.exceptions.Timeout:
            is_error = True
            logger.error(f"Request timeout: {url}")
            raise
        except requests.exceptions.RequestException as e:
            is_error = True
            logger.error(f"Request failed: {url}: {e}")
            raise
        finally:
            # Record API metrics
            latency_ms = (time.perf_counter() - start_time) * 1000
            self._record_api_call(endpoint_name, latency_ms, is_error)

    def _extract_endpoint_name(self, endpoint: str) -> str:
        """
        Extract a normalized endpoint name for metrics grouping.

        Converts paths like 'organizations/{id}/data-structures' to 'data-structures'.
        This groups API calls by logical endpoint rather than full path.
        """
        parts = endpoint.strip("/").split("/")
        # Skip organization ID parts and return the meaningful endpoint name
        meaningful_parts = [
            p for p in parts if p not in ("organizations", self.organization_id, "v1")
        ]
        if meaningful_parts:
            # Return last meaningful part (e.g., 'data-structures', 'event-specs')
            return meaningful_parts[-1]
        return "unknown"

    def _record_api_call(
        self, endpoint: str, latency_ms: float, is_error: bool
    ) -> None:
        """Record API call metrics if report is available."""
        if self.report is not None:
            self.report.record_api_call("bdp", endpoint, latency_ms, is_error)

    # ============================================
    # Organization API
    # ============================================

    def get_organization(self) -> Optional[Organization]:
        """
        Get organization details including warehouse destination configuration.

        Returns:
            Organization details with warehouse source information, or None if error
        """
        endpoint = f"organizations/{self.organization_id}"

        logger.debug("Fetching organization details from Snowplow")

        try:
            response_data = self._request("GET", endpoint)

            if not response_data:
                logger.warning("Empty response from organization endpoint")
                return None

            try:
                organization = Organization.model_validate(response_data)
                logger.info(f"Found organization: {organization.name}")

                if organization.source:
                    logger.info(
                        f"Organization has warehouse destination: {organization.source.name}"
                    )

                return organization
            except ValidationError as parse_error:
                error_msg = f"Failed to parse organization response: {parse_error}"
                logger.error(error_msg)
                if self.report:
                    self.report.report_warning("organization_parsing", error_msg)
                logger.debug(
                    f"Raw response data: {json.dumps(response_data, indent=2, default=str)}"
                )
                return None

        except (requests.exceptions.RequestException, ResourceNotFoundError) as e:
            error_msg = f"Failed to fetch organization details: {e}"
            logger.warning(error_msg)
            if self.report:
                self.report.report_warning("organization_fetch", error_msg)
            return None

    # ============================================
    # Data Structures API
    # ============================================

    def get_data_structures(
        self,
        vendor: Optional[str] = None,
        name: Optional[str] = None,
        page_size: int = 100,
    ) -> List[DataStructure]:
        """
        Get all data structures (schemas) from organization with pagination support.

        API Reference: https://docs.snowplow.io/docs/understanding-tracking-design/managing-your-data-structures/api/

        Args:
            vendor: Optional filter by schema vendor
            name: Optional filter by schema name
            page_size: Number of results per page (default: 100)

        Returns:
            List of data structures
        """
        # GET /organizations/{organizationId}/data-structures/v1
        # Supports pagination via 'from' and 'size' parameters
        endpoint = f"organizations/{self.organization_id}/data-structures/v1"

        params: Dict[str, str] = {}
        if vendor:
            params["vendor"] = vendor
        if name:
            params["name"] = name

        logger.debug(
            f"Fetching data structures from Snowplow (vendor={vendor}, name={name})"
        )

        all_structures: List[DataStructure] = []
        offset = 0

        while True:
            # Add pagination parameters
            params["from"] = str(offset)
            params["size"] = str(page_size)

            logger.debug(f"Fetching page: offset={offset}, size={page_size}")

            try:
                response_data = self._request("GET", endpoint, params=params)
            except (
                requests.exceptions.RequestException,
                ResourceNotFoundError,
            ) as e:
                error_msg = (
                    f"Failed to fetch data structures page at offset {offset}: {e}"
                )
                logger.warning(error_msg)
                if self.report:
                    self.report.report_warning("data_structures_pagination", error_msg)
                break  # Return partial results instead of losing everything

            # BDP API returns direct array (per Swagger spec)
            if not isinstance(response_data, list):
                error_msg = f"Expected list from BDP API, got {type(response_data)}"
                logger.error(error_msg)
                if self.report:
                    self.report.report_warning("data_structures_format", error_msg)
                break

            # No more results
            if not response_data:
                break

            try:
                page_structures = [
                    DataStructure.model_validate(ds) for ds in response_data
                ]
                all_structures.extend(page_structures)

                # If we got fewer results than page_size, we've reached the end
                if len(page_structures) < page_size:
                    break

                offset += len(page_structures)

            except ValidationError as e:
                error_msg = (
                    f"Failed to parse data structures page at offset {offset}: {e}"
                )
                logger.error(error_msg)
                logger.debug(
                    f"Raw response data (first item): {json.dumps(response_data[0] if response_data else 'empty', indent=2, default=str)}"
                )
                if self.report:
                    self.report.report_warning("data_structures_parsing", error_msg)
                break

        logger.info(f"Found {len(all_structures)} data structures total")
        return all_structures

    def get_data_structure(self, data_structure_hash: str) -> Optional[DataStructure]:
        """
        Get specific data structure by hash.

        API Reference: https://docs.snowplow.io/docs/understanding-tracking-design/managing-your-data-structures/api/

        Args:
            data_structure_hash: Data structure hash identifier

        Returns:
            Data structure or None if not found
        """
        # GET /organizations/{organizationId}/data-structures/v1/{dataStructureHash}
        # Docs: https://docs.snowplow.io/docs/understanding-tracking-design/managing-your-data-structures/api/
        endpoint = f"organizations/{self.organization_id}/data-structures/v1/{data_structure_hash}"

        try:
            response_data = self._request("GET", endpoint)
        except (requests.exceptions.RequestException, ResourceNotFoundError) as e:
            error_msg = f"Failed to fetch data structure {data_structure_hash}: {e}"
            logger.warning(error_msg)
            if self.report:
                self.report.report_warning("data_structure_fetch", error_msg)
            return None

        if not response_data:
            return None

        try:
            # Real BDP API returns data structure directly, not wrapped
            return DataStructure.model_validate(response_data)
        except ValidationError as e:
            error_msg = f"Failed to parse data structure {data_structure_hash}: {e}"
            logger.error(error_msg)
            if self.report:
                self.report.report_warning("data_structure_parsing", error_msg)
            logger.debug(
                f"Raw response data: {json.dumps(response_data, indent=2, default=str)}"
            )
            return None

    def get_data_structure_version(
        self, data_structure_hash: str, version: str, env: Optional[str] = None
    ) -> Optional[DataStructure]:
        """
        Get specific version of a data structure with full schema definition.

        This endpoint returns the actual JSON Schema definition with properties/fields,
        unlike the list/get endpoints which only return metadata.

        API Reference: https://docs.snowplow.io/docs/understanding-tracking-design/managing-your-data-structures/api/

        Args:
            data_structure_hash: Data structure hash identifier
            version: Schema version (e.g., "1-0-0")
            env: Optional environment filter (e.g., "PROD", "DEV")

        Returns:
            Data structure with full schema definition, or None if not found
        """
        # GET /organizations/{organizationId}/data-structures/v1/{schemaHash}/versions/{version}
        # This returns SelfDescribingSchema_Json with the actual schema definition
        endpoint = f"organizations/{self.organization_id}/data-structures/v1/{data_structure_hash}/versions/{version}"

        params = {}
        if env:
            params["env"] = env

        logger.debug(
            f"Fetching schema definition for {data_structure_hash} version {version}"
        )

        try:
            response_data = self._request("GET", endpoint, params=params)
        except (requests.exceptions.RequestException, ResourceNotFoundError) as e:
            error_msg = (
                f"Failed to fetch schema version {data_structure_hash}/{version}: {e}"
            )
            logger.warning(error_msg)
            if self.report:
                self.report.report_warning("schema_version_fetch", error_msg)
            return None

        if not response_data:
            return None

        try:
            # Response format: The JSON Schema properties are at root level, with 'self' merged in
            # Example: {"self": {...}, "properties": {...}, "type": "object", "required": [...]}
            # Not nested like {"self": {...}, "schema": {...}}

            # Extract self-descriptor info
            self_desc = response_data.get("self", {})

            if not self_desc:
                logger.warning(
                    f"Invalid schema response for {data_structure_hash}/{version}: missing 'self' field"
                )
                return None

            # Schema definition is everything except the 'self' field
            schema_def = {k: v for k, v in response_data.items() if k != "self"}

            if not schema_def:
                logger.warning(
                    f"Invalid schema response for {data_structure_hash}/{version}: no schema properties"
                )
                return None

            # Build DataStructure with the schema definition
            data_structure_dict = {
                "hash": data_structure_hash,
                "vendor": self_desc.get("vendor"),
                "name": self_desc.get("name"),
                "data": {
                    "self": self_desc,
                    **schema_def,  # Merge schema properties
                },
            }

            return DataStructure.model_validate(data_structure_dict)
        except ValidationError as e:
            error_msg = (
                f"Failed to parse schema version {data_structure_hash}/{version}: {e}"
            )
            logger.error(error_msg)
            if self.report:
                self.report.report_warning("schema_version_parsing", error_msg)
            logger.debug(
                f"Raw response data: {json.dumps(response_data, indent=2, default=str)}"
            )
            return None

    def get_data_structure_deployments(
        self, data_structure_hash: str
    ) -> List[DataStructureDeployment]:
        """
        Get deployment information for data structure.

        Returns ALL historical deployments, not just current deployments per environment.
        Uses pagination to retrieve complete deployment history.

        API Reference: https://docs.snowplow.io/docs/understanding-tracking-design/managing-your-data-structures/api/

        Args:
            data_structure_hash: Data structure hash identifier

        Returns:
            List of all deployments (historical and current)
        """
        # GET /organizations/{organizationId}/data-structures/v1/{dataStructureHash}/deployments
        # IMPORTANT: Must use pagination params (from, size) to get full deployment history
        # Without pagination, only returns CURRENT deployment per environment
        endpoint = f"organizations/{self.organization_id}/data-structures/v1/{data_structure_hash}/deployments"

        # Use large page size to get all deployments in one call
        # Typical schemas have <100 deployments, 1000 is more than sufficient
        params = {"from": 0, "size": 1000}

        try:
            response_data = self._request("GET", endpoint, params=params)
        except ResourceNotFoundError:
            logger.info(
                f"Deployments not found for data structure {data_structure_hash} (404)"
            )
            return []
        except (requests.exceptions.RequestException, PermissionError) as e:
            error_msg = f"Failed to fetch deployments for {data_structure_hash}: {e}"
            logger.warning(error_msg)
            if self.report:
                self.report.report_warning("deployment_fetch", error_msg)
            return []

        if not response_data:
            return []

        try:
            # With pagination params, response is a direct array
            # Without pagination, response is {"data": [...]}
            if isinstance(response_data, list):
                deployments_list = response_data
            else:
                deployments_list = response_data.get("data", [])

            return [DataStructureDeployment.model_validate(d) for d in deployments_list]
        except ValidationError as e:
            error_msg = f"Failed to parse deployments for {data_structure_hash}: {e}"
            logger.error(error_msg)
            if self.report:
                self.report.report_warning("deployment_parsing", error_msg)
            logger.debug(
                f"Raw response data: {json.dumps(response_data, indent=2, default=str)}"
            )
            return []

    # ============================================
    # Event Specifications API
    # ============================================

    def get_event_specifications(self) -> List[EventSpecification]:
        """
        Get all event specifications from organization.

        API Reference: https://docs.snowplow.io/docs/data-product-studio/event-specifications/api/

        Returns:
            List of event specifications
        """
        # GET /organizations/{organizationId}/event-specs/v1
        # Docs: https://docs.snowplow.io/docs/data-product-studio/event-specifications/api/
        endpoint = f"organizations/{self.organization_id}/event-specs/v1"

        logger.debug("Fetching event specifications from Snowplow")

        try:
            response_data = self._request("GET", endpoint)
        except ResourceNotFoundError:
            # Endpoint may not be available for some organizations
            logger.info(
                "Event specifications endpoint not available (404) - this is normal for some organizations"
            )
            return []

        # Handle empty response
        if not response_data:
            logger.debug("Event specifications endpoint returned empty response")
            return []

        # Event specifications API uses wrapped format (unlike data structures API)
        # Response structure: {"data": [...], "includes": [...], "errors": [...]}
        if not isinstance(response_data, dict):
            error_msg = f"Expected dict (wrapped response) from event specs API, got {type(response_data)}"
            logger.error(error_msg)
            if self.report:
                self.report.report_warning("event_specs_format", error_msg)
            return []

        try:
            response = EventSpecificationsResponse.model_validate(response_data)
            logger.info(f"Found {len(response.data)} event specifications")

            # Log any errors/warnings from API
            if response.errors:
                logger.warning(
                    f"API returned {len(response.errors)} errors/warnings for event specifications"
                )

            return response.data
        except ValidationError as e:
            error_msg = f"Failed to parse event specifications: {e}"
            logger.error(error_msg)
            if self.report:
                self.report.report_warning("event_specs_parsing", error_msg)
            logger.debug(
                f"Raw response data: {json.dumps(response_data, indent=2, default=str)}"
            )
            return []

    def get_event_specification(
        self, event_spec_id: str
    ) -> Optional[EventSpecification]:
        """
        Get specific event specification by ID.

        API Reference: https://docs.snowplow.io/docs/data-product-studio/event-specifications/api/

        Args:
            event_spec_id: Event specification ID

        Returns:
            Event specification or None if not found
        """
        # GET /organizations/{organizationId}/event-specs/v1/{eventSpecId}
        # Docs: https://docs.snowplow.io/docs/data-product-studio/event-specifications/api/
        endpoint = (
            f"organizations/{self.organization_id}/event-specs/v1/{event_spec_id}"
        )

        try:
            response_data = self._request("GET", endpoint)
        except (requests.exceptions.RequestException, ResourceNotFoundError) as e:
            error_msg = f"Failed to fetch event specification {event_spec_id}: {e}"
            logger.warning(error_msg)
            if self.report:
                self.report.report_warning("event_spec_fetch", error_msg)
            return None

        if not response_data:
            return None

        try:
            return EventSpecification.model_validate(response_data.get("data", {}))
        except ValidationError as e:
            error_msg = f"Failed to parse event specification {event_spec_id}: {e}"
            logger.error(error_msg)
            if self.report:
                self.report.report_warning("event_spec_parsing", error_msg)
            logger.debug(
                f"Raw response data: {json.dumps(response_data, indent=2, default=str)}"
            )
            return None

    # ============================================
    # Tracking Plans API
    # ============================================

    def get_tracking_plans(self) -> List[TrackingPlan]:
        """
        Get all tracking plans from organization.

        Uses the /data-products/v2 API which provides richer metadata than the
        legacy /tracking-scenarios/v1 endpoint. Snowplow originally called
        tracking plans "data products" — the v2 API still uses that path.

        Returns:
            List of tracking plans
        """
        # GET /organizations/{organizationId}/data-products/v2
        # Note: API path uses legacy 'data-products' name (Snowplow renamed to tracking plans)
        endpoint = f"organizations/{self.organization_id}/data-products/v2"

        logger.debug("Fetching tracking plans from Snowplow")

        try:
            response_data = self._request("GET", endpoint)
        except ResourceNotFoundError:
            logger.info(
                "Tracking plans endpoint not available (404) - this is normal for some organizations"
            )
            return []

        # Handle empty response
        if not response_data:
            logger.debug("Tracking plans endpoint returned empty response")
            return []

        # Tracking plans API uses wrapped format
        # Response structure: {"data": [...], "includes": {...}, "errors": [...]}
        if not isinstance(response_data, dict):
            error_msg = f"Expected dict (wrapped response) from tracking plans API, got {type(response_data)}"
            logger.error(error_msg)
            if self.report:
                self.report.report_warning("tracking_plans_format", error_msg)
            return []

        try:
            response = TrackingPlansResponse.model_validate(response_data)
            logger.info(f"Found {len(response.data)} tracking plans")

            # Log any errors/warnings from API
            if response.errors:
                logger.warning(
                    f"API returned {len(response.errors)} errors/warnings for tracking plans"
                )

            return response.data
        except ValidationError as e:
            error_msg = f"Failed to parse tracking plans: {e}"
            logger.error(error_msg)
            if self.report:
                self.report.report_warning("tracking_plans_parsing", error_msg)
            logger.debug(
                f"Raw response data: {json.dumps(response_data, indent=2, default=str)}"
            )
            return []

    def get_tracking_plan(self, plan_id: str) -> Optional[TrackingPlan]:
        """
        Get specific tracking plan by ID.

        Args:
            plan_id: Tracking plan ID

        Returns:
            Tracking plan or None if not found
        """
        # GET /organizations/{organizationId}/data-products/v2/{id}
        # Note: API path uses legacy 'data-products' name
        endpoint = f"organizations/{self.organization_id}/data-products/v2/{plan_id}"

        try:
            response_data = self._request("GET", endpoint)
        except (requests.exceptions.RequestException, ResourceNotFoundError) as e:
            error_msg = f"Failed to fetch tracking plan {plan_id}: {e}"
            logger.warning(error_msg)
            if self.report:
                self.report.report_warning("tracking_plan_fetch", error_msg)
            return None

        if not response_data:
            return None

        try:
            return TrackingPlan.model_validate(response_data.get("data", {}))
        except ValidationError as e:
            error_msg = f"Failed to parse tracking plan {plan_id}: {e}"
            logger.error(error_msg)
            if self.report:
                self.report.report_warning("tracking_plan_parsing", error_msg)
            logger.debug(
                f"Raw response data: {json.dumps(response_data, indent=2, default=str)}"
            )
            return None

    def get_data_models(self, tracking_plan_id: str) -> List[DataModel]:
        """
        Get data models for a tracking plan.

        Data models define transformations and outputs to warehouse destinations.
        Each model references a destination (warehouse) and specifies the table name.

        API Reference: https://docs.snowplow.io/docs/data-product-studio/data-models/api/

        Args:
            tracking_plan_id: Tracking plan UUID

        Returns:
            List of data models for the tracking plan
        """
        # GET /organizations/{organizationId}/data-products/v2/{id}/data-models/v1
        # Note: API path uses legacy 'data-products' name
        endpoint = f"organizations/{self.organization_id}/data-products/v2/{tracking_plan_id}/data-models/v1"

        logger.debug(f"Fetching data models for tracking plan {tracking_plan_id}")

        try:
            response_data = self._request("GET", endpoint)
        except (requests.exceptions.RequestException, ResourceNotFoundError) as e:
            error_msg = (
                f"Failed to fetch data models for tracking plan {tracking_plan_id}: {e}"
            )
            logger.warning(error_msg)
            if self.report:
                self.report.report_warning("data_models_fetch", error_msg)
            return []

        # Handle 404 or empty response
        if not response_data:
            logger.debug(
                f"No data models found for tracking plan {tracking_plan_id} (empty response)"
            )
            return []

        # Response is an array of data models
        if not isinstance(response_data, list):
            logger.warning(
                f"Expected list from data models API, got {type(response_data)}"
            )
            return []

        # Parse data models
        data_models = []
        for model_data in response_data:
            try:
                data_model = DataModel.model_validate(model_data)
                data_models.append(data_model)
            except ValidationError as e:
                logger.warning(f"Failed to parse data model: {e}")
                logger.debug(
                    f"Raw model data: {json.dumps(model_data, indent=2, default=str)}"
                )
                continue

        logger.debug(
            f"Found {len(data_models)} data models for tracking plan {tracking_plan_id}"
        )

        return data_models

    # ============================================
    # Pipeline Methods
    # ============================================

    def get_pipelines(self) -> List[Pipeline]:
        """
        Get all pipelines from organization.

        API Reference: Discovered via testing - not yet officially documented

        Returns:
            List of pipelines
        """
        # GET /organizations/{organizationId}/pipelines/v1
        endpoint = f"organizations/{self.organization_id}/pipelines/v1"

        logger.debug("Fetching pipelines from Snowplow")

        try:
            response_data = self._request("GET", endpoint)
        except ResourceNotFoundError:
            # Endpoint may not be available for some organizations
            logger.info(
                "Pipelines endpoint not available (404) - this is normal for some organizations"
            )
            return []

        # Handle empty response
        if not response_data:
            logger.debug("Pipelines endpoint returned empty response")
            return []

        try:
            response = PipelinesResponse.model_validate(response_data)
            logger.info(f"Found {len(response.pipelines)} pipelines")

            return response.pipelines
        except ValidationError as e:
            error_msg = f"Failed to parse pipelines: {e}"
            logger.error(error_msg)
            if self.report:
                self.report.report_warning("pipelines_parsing", error_msg)
            logger.debug(
                f"Raw response data: {json.dumps(response_data, indent=2, default=str)}"
            )
            return []

    def get_pipeline(self, pipeline_id: str) -> Optional[Pipeline]:
        """
        Get specific pipeline by ID.

        Args:
            pipeline_id: Pipeline ID

        Returns:
            Pipeline or None if not found
        """
        # GET /organizations/{organizationId}/pipelines/v1/{pipelineId}
        endpoint = f"organizations/{self.organization_id}/pipelines/v1/{pipeline_id}"

        try:
            response_data = self._request("GET", endpoint)
        except (requests.exceptions.RequestException, ResourceNotFoundError) as e:
            error_msg = f"Failed to fetch pipeline {pipeline_id}: {e}"
            logger.warning(error_msg)
            if self.report:
                self.report.report_warning("pipeline_fetch", error_msg)
            return None

        if not response_data:
            return None

        try:
            return Pipeline.model_validate(response_data)
        except ValidationError as e:
            error_msg = f"Failed to parse pipeline {pipeline_id}: {e}"
            logger.error(error_msg)
            if self.report:
                self.report.report_warning("pipeline_parsing", error_msg)
            logger.debug(
                f"Raw response data: {json.dumps(response_data, indent=2, default=str)}"
            )
            return None

    def get_enrichments(self, pipeline_id: str) -> List[Enrichment]:
        """
        Get all enrichments for a pipeline.

        API Reference: Discovered via testing - resources/v1 path

        Args:
            pipeline_id: Pipeline ID

        Returns:
            List of enrichments
        """
        # GET /organizations/{organizationId}/resources/v1/pipelines/{pipelineId}/configuration/enrichments
        endpoint = f"organizations/{self.organization_id}/resources/v1/pipelines/{pipeline_id}/configuration/enrichments"

        logger.debug(f"Fetching enrichments for pipeline {pipeline_id}")

        try:
            response_data = self._request("GET", endpoint)
        except ResourceNotFoundError:
            # Endpoint may not be available for some organizations
            logger.info(
                f"Enrichments endpoint not available for pipeline {pipeline_id} (404)"
            )
            return []

        # Handle empty response
        if not response_data:
            logger.debug(f"No enrichments found for pipeline {pipeline_id}")
            return []

        # Parse enrichments individually so one bad item doesn't lose all
        enrichments: List[Enrichment] = []
        for enrichment_data in response_data:
            try:
                enrichments.append(Enrichment.model_validate(enrichment_data))
            except ValidationError as e:
                error_msg = f"Failed to parse enrichment in pipeline {pipeline_id}: {e}"
                logger.warning(error_msg)
                if self.report:
                    self.report.report_warning("enrichment_parsing", error_msg)
                continue

        logger.info(f"Found {len(enrichments)} enrichments")
        return enrichments

    def get_destinations(self) -> List[Destination]:
        """
        Get all warehouse destinations (loaders) from organization.

        API Reference: /destinations/v3 endpoint (discovered via UI inspection)

        Destinations define how enriched events are loaded into warehouses,
        including database, schema, and table configuration.

        Returns:
            List of destinations
        """
        # GET /organizations/{organizationId}/destinations/v3
        endpoint = f"organizations/{self.organization_id}/destinations/v3"

        logger.debug("Fetching destinations from Snowplow")

        try:
            response_data = self._request("GET", endpoint)
        except ResourceNotFoundError:
            # Endpoint may not be available for some organizations
            logger.info(
                "Destinations endpoint not available (404) - this is normal for some organizations"
            )
            return []

        # Handle empty response
        if not response_data:
            logger.debug("Destinations endpoint returned empty response")
            return []

        # Response is a direct array of destinations
        if not isinstance(response_data, list):
            error_msg = (
                f"Expected list from destinations API, got {type(response_data)}"
            )
            logger.error(error_msg)
            if self.report:
                self.report.report_warning("destinations_format", error_msg)
            return []

        # Parse destinations individually so one bad item doesn't lose all
        destinations: List[Destination] = []
        for dest_data in response_data:
            try:
                destinations.append(Destination.model_validate(dest_data))
            except ValidationError as e:
                error_msg = f"Failed to parse destination: {e}"
                logger.warning(error_msg)
                if self.report:
                    self.report.report_warning("destination_parsing", error_msg)
                continue

        logger.info(f"Found {len(destinations)} destinations")
        return destinations

    # ============================================
    # Helper Methods
    # ============================================

    def test_connection(self) -> bool:
        """
        Test API connectivity and authentication.

        Returns:
            True if connection successful, False otherwise
        """
        try:
            # Try to fetch data structures as connectivity test
            self.get_data_structures()
            return True
        except Exception as e:
            logger.error(f"Connection test failed: {e}")
            return False

    def get_users(self) -> List[User]:
        """
        Get all users in the organization.

        API Reference: https://console.snowplowanalytics.com/api/msc/v1/docs/#/Users/getOrganizationsOrganizationidUsers

        This is used to map initiatorId from deployments to user email addresses
        for ownership extraction.

        Returns:
            List of users with id, email, and name fields
        """
        # GET /organizations/{organizationId}/users
        endpoint = f"organizations/{self.organization_id}/users"

        logger.debug("Fetching users from Snowplow BDP")

        try:
            response_data = self._request("GET", endpoint)
        except (requests.exceptions.RequestException, ResourceNotFoundError) as e:
            error_msg = f"Failed to fetch users: {e}"
            logger.warning(error_msg)
            if self.report:
                self.report.report_warning("users_fetch", error_msg)
            return []

        # BDP API returns direct array (per Swagger spec)
        if not isinstance(response_data, list):
            error_msg = f"Expected list from users API, got {type(response_data)}"
            logger.error(error_msg)
            if self.report:
                self.report.report_warning("users_format", error_msg)
            return []

        # Parse users individually so one bad item doesn't lose all
        users: List[User] = []
        for user_data in response_data:
            try:
                users.append(User.model_validate(user_data))
            except ValidationError as e:
                error_msg = f"Failed to parse user: {e}"
                logger.warning(error_msg)
                if self.report:
                    self.report.report_warning("user_parsing", error_msg)
                continue

        logger.info(f"Found {len(users)} users")
        return users
